import unittest
from io import StringIO
from unittest.mock import patch

from cubing_algs.constants import INITIAL_STATE
from cubing_algs.exceptions import InvalidCubeStateError
from cubing_algs.exceptions import InvalidFaceError
from cubing_algs.exceptions import InvalidMoveError
from cubing_algs.move import Move
from cubing_algs.parsing import parse_moves
from cubing_algs.transform.fat import unfat_rotation_moves
from cubing_algs.vcube import VCube


class VCubeTestCase(unittest.TestCase):
    maxDiff = None

    def test_state(self):
        cube = VCube()

        self.assertEqual(
            cube.state,
            INITIAL_STATE,
        )

        result = cube.rotate('R2 U2')
        self.assertEqual(
            result,
            'DUUDUUDUULLLRRRRRRFBBFFBFFBDDUDDUDDURRRLLLLLLFFBFBBFBB',
        )

        self.assertEqual(
            result,
            cube.state,
        )

    def test_is_solved(self):
        cube = VCube()

        self.assertTrue(
            cube.is_solved,
        )

        cube.rotate('R2 U2')
        self.assertFalse(
            cube.is_solved,
        )

    def test_is_solved_oriented(self):
        cube = VCube()
        cube.rotate('z2')

        self.assertTrue(cube.is_solved)

    def test_copy(self):
        cube = VCube()
        cube.rotate('R2 F2 D2 B')
        copy = cube.copy()

        self.assertEqual(
            cube.state,
            copy.state,
        )
        self.assertFalse(copy.history)

    def test_full_copy(self):
        cube = VCube()
        cube.rotate('R2 F2 D2 B')
        copy = cube.copy(full=True)

        self.assertEqual(
            cube.state,
            copy.state,
        )
        self.assertTrue(copy.history)

    def test_from_cubies(self):
        cp = [0, 5, 2, 1, 7, 4, 6, 3]
        co = [1, 2, 0, 2, 1, 1, 0, 2]
        ep = [1, 9, 2, 3, 11, 8, 6, 7, 4, 5, 10, 0]
        eo = [1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0]
        so = [0, 1, 2, 3, 4, 5]
        facelets = 'UUFUUFLLFUUURRRRRRFFRFFDFFDRRBDDBDDBLLDLLDLLDLBBUBBUBB'

        cube = VCube.from_cubies(cp, co, ep, eo, so)
        self.assertEqual(cube.state, facelets)

        cube = VCube()
        cube.rotate('F R')

        self.assertEqual(cube.state, facelets)

    def test_to_cubies(self):
        cp = [0, 5, 2, 1, 7, 4, 6, 3]
        co = [1, 2, 0, 2, 1, 1, 0, 2]
        ep = [1, 9, 2, 3, 11, 8, 6, 7, 4, 5, 10, 0]
        eo = [1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0]
        so = [0, 1, 2, 3, 4, 5]
        facelets = 'UUFUUFLLFUUURRRRRRFFRFFDFFDRRBDDBDDBLLDLLDLLDLBBUBBUBB'

        self.assertEqual(
            VCube(facelets).to_cubies,
            (
                cp, co,
                ep, eo,
                so,
            ),
        )

    def test_from_cubies_equality(self):
        cube = VCube()
        cube.rotate('F R')
        n_cube = VCube.from_cubies(*cube.to_cubies)

        self.assertEqual(
            cube.state,
            n_cube.state,
        )

    def test_from_cubies_oriented_equality(self):
        cube = VCube()
        cube.rotate('F R x')
        n_cube = VCube.from_cubies(*cube.to_cubies)

        self.assertEqual(
            cube.state,
            n_cube.state,
        )

    def test_display(self):
        cube = VCube()
        cube.rotate('F R U')

        result = cube.display()

        lines = [line for line in result.split('\n') if line.strip()]

        self.assertEqual(len(lines), 9)
        self.assertEqual(len(cube.history), 3)

    def test_display_orientation_restore(self):
        cube = VCube()
        cube.rotate('F R U')

        self.assertEqual(len(cube.history), 3)

        state = cube.state

        cube.display(orientation='DF')

        self.assertEqual(len(cube.history), 3)
        self.assertEqual(state, cube.state)

    def test_display_orientation_different(self):
        cube_1 = VCube()
        cube_2 = VCube()

        view_1 = cube_1.display()
        view_2 = cube_2.display(orientation='DF')

        self.assertNotEqual(view_1, view_2)

    def test_get_face(self):
        cube = VCube()
        cube.rotate('F R U')

        self.assertEqual(
            cube.get_face('U'),
            'LUULUUFFF',
        )

        cube.rotate('z2')

        self.assertEqual(
            cube.get_face('U'),
            'BDDBDDBRR',
        )

    def test_get_face_by_center(self):
        cube = VCube()
        cube.rotate('F R U')

        self.assertEqual(
            cube.get_face_by_center('U'),
            'LUULUUFFF',
        )

        cube.rotate('z2')

        self.assertEqual(
            cube.get_face_by_center('U'),
            'FFFUULUUL',
        )

    def test_get_face_center(self):
        cube = VCube()
        cube.rotate('F R U')

        self.assertEqual(
            cube.get_face_by_center('U'),
            'LUULUUFFF',
        )

        cube.rotate('z2')

        self.assertEqual(
            cube.get_face_by_center('U'),
            'FFFUULUUL',
        )

    def test_get_face_index(self):
        cube = VCube()
        cube.rotate('F R U')

        self.assertEqual(
            cube.get_face_index('U'),
            0,
        )

        cube.rotate('z2')

        self.assertEqual(
            cube.get_face_index('U'),
            3,
        )

    def test_get_face_center_indexes(self):
        cube = VCube()
        cube.rotate('F R U')

        self.assertEqual(
            cube.get_face_center_indexes(),
            ['U', 'R', 'F', 'D', 'L', 'B'],
        )

        cube.rotate('z2')

        self.assertEqual(
            cube.get_face_center_indexes(),
            ['D', 'L', 'F', 'U', 'R', 'B'],
        )

    def test_str(self):
        cube = VCube()
        cube.rotate('F R U')

        self.assertEqual(
            str(cube),
            'U: LUULUUFFF\n'
            'R: LBBRRRRRR\n'
            'F: UUUFFDFFD\n'
            'D: RRBDDBDDB\n'
            'L: FFRLLDLLD\n'
            'B: LLDUBBUBB',
        )

    def test_repr(self):
        cube = VCube()
        cube.rotate('F R U')

        self.assertEqual(
            repr(cube),
            "VCube('LUULUUFFFLBBRRRRRRUUUFFDFFDRRBDDBDDBFFRLLDLLDLLDUBBUBB')",
        )


class VCubeOrientedCopyTestCase(unittest.TestCase):
    maxDiff = None

    def test_oriented_copy_faces(self):
        cube = VCube()

        self.assertNotEqual(
            cube.state,
            cube.oriented_copy('DF').state,
        )

    def test_oriented_copy_top_only(self):
        cube = VCube()

        self.assertNotEqual(
            cube.state,
            cube.oriented_copy('D').state,
        )

    def test_oriented_copy_faces_stable(self):
        cube = VCube()
        base_state = cube.state
        cube.oriented_copy('UF')

        self.assertEqual(
            cube.state,
            base_state,
        )

    def test_oriented_copy_invalid_empty(self):
        cube = VCube()

        with self.assertRaises(InvalidFaceError):
            cube.oriented_copy('')

    def test_oriented_copy_invalid_too_much(self):
        cube = VCube()

        with self.assertRaises(InvalidFaceError):
            cube.oriented_copy('FRU')

    def test_oriented_copy_invalid_top_face(self):
        cube = VCube()

        with self.assertRaises(InvalidFaceError):
            cube.oriented_copy('TF')

    def test_oriented_copy_invalid_front_face(self):
        cube = VCube()

        with self.assertRaises(InvalidFaceError):
            cube.oriented_copy('FT')

    def test_oriented_copy_invalid_opposite_face(self):
        cube = VCube()

        with self.assertRaises(InvalidFaceError):
            cube.oriented_copy('FB')

    def test_history_preservation(self):
        cube = VCube()
        cube.rotate('R F')

        self.assertEqual(
            len(cube.history),
            2,
        )

        oriented = cube.oriented_copy('DF')

        self.assertEqual(
            len(cube.history),
            2,
        )

        self.assertEqual(
            len(oriented.history),
            0,
        )

    def test_history_tracking(self):
        cube = VCube()
        cube.rotate('R F')

        oriented = cube.oriented_copy('DF', full=True)

        self.assertEqual(
            len(oriented.history),
            3,
        )

        oriented = cube.oriented_copy('DB', full=True)

        self.assertEqual(
            len(oriented.history),
            4,
        )

    def test_all_reorientation(self):
        orientations = [
            'UF', 'UB', 'UR', 'UL',
            'DF', 'DB', 'DR', 'DL',
            'FU', 'FD', 'FR', 'FL',
            'BU', 'BD', 'BR', 'BL',
            'RU', 'RD', 'RF', 'RB',
            'LU', 'LD', 'LF', 'LB',
        ]

        for orientation in orientations:
            with self.subTest(orientation=orientation):
                cube = VCube().oriented_copy(orientation)

                self.assertEqual(
                    cube.state[4],
                    orientation[0],
                )

                self.assertEqual(
                    cube.state[21],
                    orientation[1],
                )


class VCubeCheckIntegrityTestCase(unittest.TestCase):
    """Tests pour les nouvelles vérifications de check_integrity"""

    def test_initial(self):
        initial = 'DUUDUUDUULLLRRRRRRFBBFFBFFBDDUDDUDDURRRLLLLLLFFBFBBFBB'

        cube = VCube(initial)

        self.assertEqual(
            cube.state,
            initial,
        )

    def test_invalid_length_no_check(self):
        initial = 'DUUDUUDUULLLRRRRRRFBBFFBFFBDDUDDUDDURRRLLLLLLFFBFBBFB'

        cube = VCube(initial, check=False)
        self.assertEqual(cube.state, initial)

    def test_invalid_length(self):
        initial = 'DUUDUUDUULLLRRRRRRFBBFFBFFBDDUDDUDDURRRLLLLLLFFBFBBFB'

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'State string must be 54 characters long',
        ):
            VCube(initial)

    def test_invalid_character(self):
        initial = 'DUUDUUDUULLLRRRRRRFBBFFBFFBDDUDDUDDURRRLLLLLLFFBFBBFBT'

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'State string can only contains U R F D L B characters',
        ):
            VCube(initial)

    def test_invalid_face(self):
        initial = 'DUUDUUDUULLLRRRRRRFBBFFBFFBDDUDDUDDURRRLLLLLLFFBFBBFBF'

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'State string must have 9 of each color',
        ):
            VCube(initial)

    def test_invalid_centers_not_unique(self):
        invalid_state = (
            'UUUUUUUUR'
            'RRRRURRRR'
            'FFFFFFFFF'
            'DDDDDDDDD'
            'LLLLLLLLL'
            'BBBBBBBBB'
        )

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Face centers must be unique',
        ):
            VCube(invalid_state)

    def test_invalid_corner_orientation_sum(self):
        co = [1, 0, 0, 0, 0, 0, 0, 0]

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Sum of corner orientations must be divisible by 3',
        ):
            VCube().check_corner_sum(co)

    def test_invalid_edge_orientation_sum(self):
        eo = [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Sum of edge orientations must be even',
        ):
            VCube().check_edge_sum(eo)

    def test_invalid_corner_permutation_duplicate(self):
        cp = [0, 0, 2, 3, 4, 5, 6, 7]

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Corner permutation must contain exactly '
                'one instance of each corner',
        ):
            VCube().check_corner_permutations(cp)

    def test_invalid_edge_permutation_duplicate(self):
        ep = [0, 0, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Edge permutation must contain exactly '
                'one instance of each edge',
        ):
            VCube().check_edge_permutations(ep)

    def test_invalid_corner_orientation_value(self):
        co = [3, 0, 0, 0, 0, 0, 0, 0]

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Corner orientation must be 0, 1, or 2 '
                'for each corner',
        ):
            VCube().check_corner_orientations(co)

    def test_invalid_edge_orientation_value(self):
        eo = [2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Edge orientation must be 0 or 1 '
                'for each edge',
        ):
            VCube().check_edge_orientations(eo)

    def test_invalid_center_orientation_value(self):
        so = [7, 0, 0, 0, 0, 0]

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Center orientation must be between 0 and 5 '
                'for each center',
        ):
            VCube().check_center_orientations(so)

    def test_invalid_permutation_parity(self):
        # Swap 0,1 = 1 inversion (odd)
        cp = [1, 0, 2, 3, 4, 5, 6, 7]
        # Identity = 0 inversions (even)
        ep = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Corner and edge permutation parities must be equal',
        ):
            VCube().check_permutation_parity(cp, ep)

    @unittest.mock.patch.object(VCube, 'check_colors')
    def test_invalid_corner_same_colors(self, *_):
        invalid_state = list(INITIAL_STATE)
        # Corner URF: same color on the 2 faces
        invalid_state[8] = invalid_state[9]
        invalid_state = ''.join(invalid_state)

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Corner 0 must have 3 different colors, got',
        ):
            VCube(invalid_state)

    @unittest.mock.patch.object(VCube, 'check_colors')
    def test_invalid_edge_same_colors(self, *_):
        invalid_state = list(INITIAL_STATE)
        # Edge UR: same color on the 2 faces
        invalid_state[5] = invalid_state[10]
        invalid_state = ''.join(invalid_state)

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Edge 0 must have 2 different colors, got ',
        ):
            VCube(invalid_state)

    @unittest.mock.patch.object(VCube, 'check_colors')
    def test_invalid_corner_opposite_colors(self, *_):
        invalid_state = list(INITIAL_STATE)
        invalid_state[8] = 'U'  # Face U
        invalid_state[9] = 'D'  # Opposite face D
        invalid_state[20] = 'F'  # Third face
        invalid_state = ''.join(invalid_state)

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Corner 0 cannot have opposite colors '
                'U and D',
        ):
            VCube(invalid_state)

    @unittest.mock.patch.object(VCube, 'check_colors')
    def test_invalid_edge_opposite_colors(self, *_):
        invalid_state = list(INITIAL_STATE)
        invalid_state[5] = 'F'
        invalid_state[10] = 'B'  # Opposite color
        invalid_state = ''.join(invalid_state)

        with self.assertRaisesRegex(
                InvalidCubeStateError,
                'Edge 0 cannot have opposite colors '
                'F and B',
        ):
            VCube(invalid_state)

    def test_valid_complex_scramble(self):
        cube = VCube()
        complex_scramble = (
            "R U2 R' D' R U' R' D R' U "
            "R U' R' U R U2 R' U' R U' R'"
        )
        cube.rotate(complex_scramble)

        self.assertTrue(cube.check_integrity())

    def test_rotations_preserve_validity(self):
        cube = VCube()
        rotations = ['x', 'y', 'z', "x'", "y'", "z'", 'x2', 'y2', 'z2']

        for rotation in rotations:
            with self.subTest(rotation=rotation):
                cube_copy = cube.copy()
                cube_copy.rotate(rotation)
                self.assertTrue(cube_copy.check_integrity())

    def test_preserve_validity(self):
        cube = VCube()

        self.assertTrue(cube.check_integrity())

    def test_oriented_preserve_validity(self):
        cube = VCube()
        cube.rotate('z2')

        self.assertTrue(
            VCube(cube.state).check_integrity(),
        )


class VCubeRotateTestCase(unittest.TestCase):

    def test_rotate_types(self):
        cube = VCube()

        self.assertEqual(
            cube.rotate(parse_moves('R F') + 'z2'),
            'BDDBDDRRRBLLDLLDLLDDDFFFFFFLLLFUUFUURRFRRURRUBBUBBUBBU',
        )

        cube = VCube()

        self.assertEqual(
            cube.rotate('z' + parse_moves('R F')),
            'LLFLLFDDDLUULUUFUUFFFFFFRRRUUURRBRRBDDRDDRDDBLBBLBBLBB',
        )

        cube = VCube()

        self.assertEqual(
            cube.rotate('z2' + parse_moves('R F')),
            'DDFDDFRRRDLLDLLFLLFFFFFFUUULLLUUBUUBRRURRURRBDBBDBBDBB',
        )

    def test_rotate_u(self):
        cube = VCube()

        self.assertEqual(
            cube.rotate('U'),
            'UUUUUUUUUBBBRRRRRRRRRFFFFFFDDDDDDDDDFFFLLLLLLLLLBBBBBB',
        )

        self.assertEqual(
            cube.rotate("U'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('U2'),
            'UUUUUUUUULLLRRRRRRBBBFFFFFFDDDDDDDDDRRRLLLLLLFFFBBBBBB',
        )

    def test_rotate_r(self):
        cube = VCube()

        self.assertEqual(
            cube.rotate('R'),
            'UUFUUFUUFRRRRRRRRRFFDFFDFFDDDBDDBDDBLLLLLLLLLUBBUBBUBB',
        )

        self.assertEqual(
            cube.rotate("R'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('R2'),
            'UUDUUDUUDRRRRRRRRRFFBFFBFFBDDUDDUDDULLLLLLLLLFBBFBBFBB',
        )

    def test_rotate_f(self):
        cube = VCube()

        self.assertEqual(
            cube.rotate('F'),
            'UUUUUULLLURRURRURRFFFFFFFFFRRRDDDDDDLLDLLDLLDBBBBBBBBB',
        )

        self.assertEqual(
            cube.rotate("F'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('F2'),
            'UUUUUUDDDLRRLRRLRRFFFFFFFFFUUUDDDDDDLLRLLRLLRBBBBBBBBB',
        )

    def test_rotate_d(self):
        cube = VCube()

        self.assertEqual(
            cube.rotate('D'),
            'UUUUUUUUURRRRRRFFFFFFFFFLLLDDDDDDDDDLLLLLLBBBBBBBBBRRR',
        )

        self.assertEqual(
            cube.rotate("D'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('D2'),
            'UUUUUUUUURRRRRRLLLFFFFFFBBBDDDDDDDDDLLLLLLRRRBBBBBBFFF',
        )

    def test_rotate_l(self):
        cube = VCube()

        self.assertEqual(
            cube.rotate('L'),
            'BUUBUUBUURRRRRRRRRUFFUFFUFFFDDFDDFDDLLLLLLLLLBBDBBDBBD',
        )

        self.assertEqual(
            cube.rotate("L'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('L2'),
            'DUUDUUDUURRRRRRRRRBFFBFFBFFUDDUDDUDDLLLLLLLLLBBFBBFBBF',
        )

    def test_rotate_b(self):
        cube = VCube()

        self.assertEqual(
            cube.rotate('B'),
            'RRRUUUUUURRDRRDRRDFFFFFFFFFDDDDDDLLLULLULLULLBBBBBBBBB',
        )

        self.assertEqual(
            cube.rotate("B'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('B2'),
            'DDDUUUUUURRLRRLRRLFFFFFFFFFDDDDDDUUURLLRLLRLLBBBBBBBBB',
        )

    def test_rotate_m(self):
        cube = VCube()

        self.assertEqual(
            cube.rotate('M'),
            'UBUUBUUBURRRRRRRRRFUFFUFFUFDFDDFDDFDLLLLLLLLLBDBBDBBDB',
        )

        self.assertEqual(
            cube.rotate("M'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('M2'),
            'UDUUDUUDURRRRRRRRRFBFFBFFBFDUDDUDDUDLLLLLLLLLBFBBFBBFB',
        )

    def test_rotate_s(self):
        cube = VCube()

        self.assertEqual(
            cube.rotate('S'),
            'UUULLLUUURURRURRURFFFFFFFFFDDDRRRDDDLDLLDLLDLBBBBBBBBB',
        )

        self.assertEqual(
            cube.rotate("S'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('S2'),
            'UUUDDDUUURLRRLRRLRFFFFFFFFFDDDUUUDDDLRLLRLLRLBBBBBBBBB',
        )

    def test_rotate_e(self):
        cube = VCube()

        self.assertEqual(
            cube.rotate('E'),
            'UUUUUUUUURRRFFFRRRFFFLLLFFFDDDDDDDDDLLLBBBLLLBBBRRRBBB',
        )

        self.assertEqual(
            cube.rotate("E'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('E2'),
            'UUUUUUUUURRRLLLRRRFFFBBBFFFDDDDDDDDDLLLRRRLLLBBBFFFBBB',
        )

    def test_rotate_x(self):
        cube = VCube()

        self.assertEqual(
            cube.rotate('x'),
            'FFFFFFFFFRRRRRRRRRDDDDDDDDDBBBBBBBBBLLLLLLLLLUUUUUUUUU',
        )

        self.assertEqual(
            cube.rotate("x'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('x2'),
            'DDDDDDDDDRRRRRRRRRBBBBBBBBBUUUUUUUUULLLLLLLLLFFFFFFFFF',
        )

    def test_rotate_y(self):
        cube = VCube()

        self.assertEqual(
            cube.rotate('y'),
            'UUUUUUUUUBBBBBBBBBRRRRRRRRRDDDDDDDDDFFFFFFFFFLLLLLLLLL',
        )

        self.assertEqual(
            cube.rotate("y'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('y2'),
            'UUUUUUUUULLLLLLLLLBBBBBBBBBDDDDDDDDDRRRRRRRRRFFFFFFFFF',
        )

    def test_rotate_z(self):
        cube = VCube()

        self.assertEqual(
            cube.rotate('z'),
            'LLLLLLLLLUUUUUUUUUFFFFFFFFFRRRRRRRRRDDDDDDDDDBBBBBBBBB',
        )

        self.assertEqual(
            cube.rotate("z'"),
            INITIAL_STATE,
        )

        self.assertEqual(
            cube.rotate('z2'),
            'DDDDDDDDDLLLLLLLLLFFFFFFFFFUUUUUUUUURRRRRRRRRBBBBBBBBB',
        )

    def test_rotate_invalid_modifier(self):
        cube = VCube()

        with self.assertRaises(InvalidMoveError):
            cube.rotate('z3')

    def test_rotate_invalid_move(self):
        cube = VCube()

        with self.assertRaises(InvalidMoveError):
            cube.rotate('T2')

    def test_real_case(self):
        cube = VCube()
        scramble = "U2 D2 F U2 F2 U R' L U2 R2 U' B2 D R2 L2 F2 U' L2 D F2 U'"

        self.assertEqual(
            cube.rotate(scramble),
            'FBFUUDUUDBFUFRLRRRLRLLFRRDBFBUBDBFUDRFBRLFLLULUDDBDBLD',
        )

    def test_real_case_2(self):
        cube = VCube()
        scramble = "F R' F' U' D2 B' L F U' F L' U F2 U' F2 B2 L2 D2 B2 D' L2"

        self.assertEqual(
            cube.rotate(scramble),
            'LDBRUUBBDFLUFRLBDDLURLFDFRLLFUFDRFDBFUDBLBRUURBDFBRRLU',
        )

    def test_real_case_3(self):
        cube = VCube()
        scramble = "F R F' U' D2 B' L F U' F L' U F2 U' F2 B2 L2 D2 B2 D' L2 B'"

        self.assertEqual(
            cube.rotate(scramble),
            'UFFRUUBBDFLLFRDBUFLURLFDBRLDFUBDRLLRBDDDLBFRRDURBBLUFU',
        )

    def test_real_case_with_algorithm(self):
        cube = VCube()
        scramble = parse_moves(
            "U2 D2 F U2 F2 U R' L U2 R2 U' B2 D R2 L2 F2 U' L2 D F2 U'",
        )

        self.assertEqual(
            cube.rotate(scramble),
            'FBFUUDUUDBFUFRLRRRLRLLFRRDBFBUBDBFUDRFBRLFLLULUDDBDBLD',
        )


class VCubeRotateWideTestCase(unittest.TestCase):

    def check_rotate(self, raw_move):
        base_move = Move(raw_move)

        for move, name in zip(
                [base_move, base_move.inverted, base_move.doubled],
                ['Base', 'Inverted', 'Doubled'],
                strict=True,
        ):
            with self.subTest(name, move=move):
                cube = VCube()
                cube_wide = VCube()

                self.assertEqual(
                    cube.rotate(move),
                    cube_wide.rotate(
                        parse_moves(
                            str(move),
                        ).transform(
                            unfat_rotation_moves,
                        ),
                    ),
                )

    def test_rotate_u(self):
        self.check_rotate('u')

    def test_rotate_r(self):
        self.check_rotate('r')

    def test_rotate_f(self):
        self.check_rotate('f')

    def test_rotate_d(self):
        self.check_rotate('d')

    def test_rotate_l(self):
        self.check_rotate('l')

    def test_rotate_b(self):
        self.check_rotate('b')


class TestVCubeShow(unittest.TestCase):

    def setUp(self):
        self.cube = VCube()

    def test_show_default_parameters(self):
        captured_output = StringIO()
        with patch('sys.stdout', captured_output):
            self.cube.show()

        output = captured_output.getvalue()

        self.assertIsInstance(output, str)
        self.assertGreater(len(output), 0)

    def test_show_with_orientation(self):
        orientations = ['', 'DF', 'FR']

        for orientation in orientations:
            with self.subTest(orientation=orientation):
                captured_output = StringIO()
                with patch('sys.stdout', captured_output):
                    self.cube.show(orientation=orientation)

                output = captured_output.getvalue()
                self.assertIsInstance(output, str)
                self.assertGreater(len(output), 0)

    def test_show_with_mode(self):
        modes = ['f2l', 'oll', 'pll']

        for mode in modes:
            with self.subTest(mode=mode):
                captured_output = StringIO()
                with patch('sys.stdout', captured_output):
                    self.cube.show(mode=mode)

                output = captured_output.getvalue()
                self.assertIsInstance(output, str)
                self.assertGreater(len(output), 0)

    def test_show_scrambled_cube(self):
        self.cube.rotate("R U R' U'")

        captured_output = StringIO()
        with patch('sys.stdout', captured_output):
            self.cube.show()

        output = captured_output.getvalue()
        self.assertIsInstance(output, str)
        self.assertGreater(len(output), 0)

        face_letters = ['U', 'R', 'F', 'D', 'L', 'B']
        for letter in face_letters:
            self.assertEqual(output.count(letter), 9)

    def test_show_output_consistency(self):
        captured_output1 = StringIO()
        with patch('sys.stdout', captured_output1):
            self.cube.show()
        output1 = captured_output1.getvalue()

        captured_output2 = StringIO()
        with patch('sys.stdout', captured_output2):
            self.cube.show()
        output2 = captured_output2.getvalue()

        self.assertEqual(output1, output2)

    def test_show_vs_display_consistency(self):
        display_result = self.cube.display()

        captured_output = StringIO()
        with patch('sys.stdout', captured_output):
            self.cube.show()
        show_result = captured_output.getvalue()

        self.assertEqual(display_result, show_result)

    def test_show_empty_parameters(self):
        captured_output = StringIO()
        with patch('sys.stdout', captured_output):
            self.cube.show(orientation='')

        output = captured_output.getvalue()
        self.assertIsInstance(output, str)
        self.assertGreater(len(output), 0)



import unittest

from cubing_algs.move import Move
from cubing_algs.parsing import parse_moves
from cubing_algs.transform.degrip import degrip_full_moves
from cubing_algs.transform.fat import refat
from cubing_algs.transform.fat import refat_moves
from cubing_algs.transform.fat import unfat_rotation_moves
from cubing_algs.transform.fat import unfat_slice_moves
from cubing_algs.transform.rotation import remove_final_rotations


class TransformFatTestCase(unittest.TestCase):

    def test_unfat_rotation_moves(self):
        provide = parse_moves('f r u')
        expect = parse_moves('B z L x D y')

        result = unfat_rotation_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_unfat_rotation_moves_part_two(self):
        provide = parse_moves('b l d')
        expect = parse_moves("F z' R x' U y'")

        result = unfat_rotation_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_unfat_rotation_moves_part_three(self):
        provide = parse_moves('r F u b')
        expect = parse_moves("L x F D y F z'")

        result = unfat_rotation_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_unfat_rotation_moves_cleaned(self):
        provide = parse_moves('f r u')
        expect = parse_moves('B D B')

        result = remove_final_rotations(
            degrip_full_moves(
                unfat_rotation_moves(
                    provide,
                ),
            ),
        )

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_unfat_rotation_moves_cleaned_part_two(self):
        provide = parse_moves('b l d')
        expect = parse_moves('F D B')

        result = remove_final_rotations(
            degrip_full_moves(
                unfat_rotation_moves(
                    provide,
                ),
            ),
        )

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_unfat_slice_moves(self):
        provide = parse_moves('f r u')
        expect = parse_moves("F S R M' U E'")

        result = unfat_slice_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_unfat_slice_moves_part_two(self):
        provide = parse_moves('b l d')
        expect = parse_moves("B S' L M D E")

        result = unfat_slice_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_unfat_slice_moves_part_three(self):
        provide = parse_moves('r F u b')
        expect = parse_moves("R M' F U E' B S'")

        result = unfat_slice_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_unfat_timed_moves(self):
        provide = parse_moves('f@1 r@2 u@3')
        expect = parse_moves('B@1 z@1 L@2 x@2 D@3 y@3')

        result = unfat_rotation_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_unfat_timed_pauses(self):
        provide = parse_moves('f@1 .@2 r@3 u@4')
        expect = parse_moves('B@1 z@1 .@2 L@3 x@3 D@4 y@4')

        result = unfat_rotation_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_refat_moves(self):
        provide = parse_moves('L x')
        expect = parse_moves('r')

        result = refat_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_refat_moves_alt(self):
        provide = parse_moves('x L')
        expect = parse_moves('r')

        result = refat_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_refat_moves_mixed(self):
        provide = parse_moves('L x f')
        expect = parse_moves('r f')

        result = refat_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_refat_moves_mixed_big_moves(self):
        provide = parse_moves('L x 2F')
        expect = parse_moves('r 2F')

        result = refat_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

        provide = parse_moves('2L x 2F')
        expect = parse_moves('2L x 2F')

        result = refat_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_refat_moves_mixed_timed_moves(self):
        provide = parse_moves('L@1 x@2 F@3')
        expect = parse_moves('r@1 F@3')

        result = refat_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

        provide = parse_moves("L'@1 x'@2 F@3")
        expect = parse_moves("r'@1 F@3")

        result = refat_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_refat_moves_mixed_timed_moves_timed_pauses(self):
        provide = parse_moves('L@1 x@2 .@3 F@4')
        expect = parse_moves('r@1 .@3 F@4')

        result = refat_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

        provide = parse_moves('L@1 x@2 F@3 .@4')
        expect = parse_moves('r@1 F@3 .@4')

        result = refat_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_refat_max(self):
        provide = parse_moves('L x')

        self.assertEqual(
            refat(provide, {}, 0),
            provide,
        )



import math
import re
from random import choice
from random import randint

from cubing_algs.algorithm import Algorithm
from cubing_algs.constants import FACE_ORDER
from cubing_algs.constants import ITERATIONS_BY_CUBE_SIZE
from cubing_algs.constants import OPPOSITE_FACES
from cubing_algs.constants import OUTER_BASIC_MOVES
from cubing_algs.parsing import parse_moves

FACE_REGEXP = re.compile(rf"({ '|'.join(FACE_ORDER) })")

MOVES_EASY_CROSS = [
    'F',
    'R',
    'B',
    'L',
]

EXCLUDE_ODD_FACES = {'D', 'L', 'B'}


def build_cube_move_set(cube_size: int, *,
                        inner_layers: bool = False) -> list[str]:
    moves = []

    for face in OUTER_BASIC_MOVES:
        moves.extend(
            [
                face,
                f"{ face }'",
                f'{ face }2',
            ],
        )
        if cube_size > 3:
            center_ceil = math.ceil(cube_size / 2)
            center_floor = math.floor(cube_size / 2)
            odd_cube = bool(cube_size % 2)
            even_cube = not odd_cube

            if cube_size > 4 or face not in EXCLUDE_ODD_FACES:
                moves.extend(
                    [
                        f'{ face }w',
                        f"{ face }w'",
                        f'{ face }w2',
                    ],
                )

            for i in range(3, center_floor + 1):
                if (
                        even_cube
                        and face in EXCLUDE_ODD_FACES
                        and i == center_floor
                ):
                    continue
                moves.extend(
                    [
                        f'{ i }{ face }w',
                        f"{ i }{ face }w'",
                        f'{ i }{ face }w2',
                    ],
                )

            if inner_layers:
                for i in range(2, center_ceil + 1):
                    if (
                            odd_cube
                            and face in EXCLUDE_ODD_FACES
                            and i == center_ceil
                    ):
                        continue

                    moves.extend(
                        [
                            f'{ i }{ face }',
                            f"{ i }{ face }'",
                            f'{ i }{ face }2',
                        ],
                    )

    return moves


def is_valid_next_move(current: str, previous: str) -> bool:
    current_move_search = FACE_REGEXP.search(current)
    previous_move_search = FACE_REGEXP.search(previous)

    if not current_move_search or not previous_move_search:
        return False

    current_move = current_move_search[0]
    previous_move = previous_move_search[0]

    if current_move == previous_move:
        return False

    return OPPOSITE_FACES[current_move] != previous_move


def random_moves(cube_size: int,
                 move_set: list[str],
                 iterations: int = 0) -> Algorithm:
    value = choice(move_set)
    moves = [value]
    previous = value

    if not iterations:
        iterations_range = ITERATIONS_BY_CUBE_SIZE[min(cube_size, 7)]
        iterations = randint(*iterations_range)

    while len(moves) < iterations:
        while not is_valid_next_move(value, previous):
            value = choice(move_set)

        previous = value
        moves.append(value)

    return parse_moves(moves)


def scramble(cube_size: int, iterations: int = 0, *,
             inner_layers: bool = False) -> Algorithm:
    move_set = build_cube_move_set(
        cube_size,
        inner_layers=inner_layers,
    )

    return random_moves(cube_size, move_set, iterations)


def scramble_easy_cross() -> Algorithm:
    return random_moves(3, MOVES_EASY_CROSS, 10)



# ruff: noqa: PLR6301
from cubing_algs.constants import CORNER_FACELET_MAP
from cubing_algs.constants import EDGE_FACELET_MAP
from cubing_algs.constants import FACE_ORDER
from cubing_algs.constants import OPPOSITE_FACES
from cubing_algs.exceptions import InvalidCubeStateError
from cubing_algs.exceptions import InvalidFaceError
from cubing_algs.facelets import facelets_to_cubies

CORNER_NUMBER = 8
CORNER_VALID_ORIENTATIONS = {0, 1, 2}

EDGE_NUMBER = 12
EDGE_VALID_ORIENTATIONS = {0, 1}


def count_inversions(permutation: list[int]) -> int:
    inversions = 0

    for i in range(len(permutation)):
        for j in range(i + 1, len(permutation)):
            if permutation[i] > permutation[j]:
                inversions += 1

    return inversions


class VCubeIntegrityChecker:
    """
    Check integrity of VCube
    """
    size: int
    face_size: int
    face_number: int

    _state: str

    def check_integrity(self) -> bool:
        self.check_length()

        color_counts: dict[str, int] = {}
        for i in self._state:
            color_counts.setdefault(i, 0)
            color_counts[i] += 1

        self.check_characters(color_counts)
        self.check_colors(color_counts)
        self.check_centers()

        cp, co, ep, eo, so = facelets_to_cubies(self._state)

        self.check_corner_permutations(cp)
        self.check_corner_orientations(co)
        self.check_corner_sum(co)
        self.check_corner_colors(cp, co)

        self.check_edge_permutations(ep)
        self.check_edge_orientations(eo)
        self.check_edge_sum(eo)
        self.check_edge_colors(ep, eo)

        self.check_permutation_parity(cp, ep)

        self.check_center_orientations(so)

        return True

    def check_length(self) -> None:
        expected_length = self.face_number * self.face_size

        if len(self._state) != expected_length:
            msg = f'State string must be { expected_length } characters long'
            raise InvalidCubeStateError(msg)

    def check_characters(self, color_counts: dict[str, int]) -> None:
        if set(color_counts.keys()) - set(FACE_ORDER):
            msg = (
                'State string can only '
                f'contains { " ".join(FACE_ORDER) } characters'
            )
            raise InvalidCubeStateError(msg)

    def check_colors(self, color_counts: dict[str, int]) -> None:
        if not all(count == self.face_size for count in color_counts.values()):
            msg = f'State string must have { self.face_size } of each color'
            raise InvalidCubeStateError(msg)

    def check_centers(self) -> None:
        actual_centers = set(self.get_face_center_indexes())

        if len(actual_centers) != self.face_number:
            msg = 'Face centers must be unique'
            raise InvalidCubeStateError(msg)

    def check_corner_permutations(self, cp: list[int]) -> None:
        if len(cp) != CORNER_NUMBER or set(cp) != set(range(CORNER_NUMBER)):
            msg = (
                'Corner permutation must contain exactly '
                'one instance of each corner (0-7)'
            )
            raise InvalidCubeStateError(msg)

    def check_corner_orientations(self, co: list[int]) -> None:
        if len(co) != CORNER_NUMBER or any(
                orientation not in CORNER_VALID_ORIENTATIONS
                for orientation in co
        ):
            msg = 'Corner orientation must be 0, 1, or 2 for each corner'
            raise InvalidCubeStateError(msg)

    def check_corner_sum(self, co: list[int]) -> None:
        if sum(co) % 3:
            msg = 'Sum of corner orientations must be divisible by 3'
            raise InvalidCubeStateError(msg)

    def check_corner_colors(self, cp: list[int], co: list[int]) -> None:
        for i, (corner_pos, _corner_ori) in enumerate(zip(cp, co, strict=True)):
            corner_facelets = [
                self._state[facelet]
                for facelet in CORNER_FACELET_MAP[corner_pos]
            ]
            if len(set(corner_facelets)) != 3:
                msg = (
                    f'Corner { i } must have 3 different colors, '
                    f'got { corner_facelets }'
                )
                raise InvalidCubeStateError(msg)

            for j, color1 in enumerate(corner_facelets):
                for _, color2 in enumerate(corner_facelets[j + 1:], j + 1):
                    if (
                            color1 in OPPOSITE_FACES
                            and OPPOSITE_FACES[color1] == color2
                    ):
                        msg = (
                            f'Corner { i } cannot have opposite colors '
                            f'{ color1 } and { color2 }'
                        )
                        raise InvalidCubeStateError(msg)

    def check_edge_permutations(self, ep: list[int]) -> None:
        if len(ep) != EDGE_NUMBER or set(ep) != set(range(EDGE_NUMBER)):
            msg = (
                'Edge permutation must contain exactly '
                'one instance of each edge (0-11)'
            )
            raise InvalidCubeStateError(msg)

    def check_edge_orientations(self, eo: list[int]) -> None:
        if len(eo) != EDGE_NUMBER or any(
                orientation not in EDGE_VALID_ORIENTATIONS
                for orientation in eo
        ):
            msg = 'Edge orientation must be 0 or 1 for each edge'
            raise InvalidCubeStateError(msg)

    def check_edge_sum(self, eo: list[int]) -> None:
        if sum(eo) % 2:
            msg = 'Sum of edge orientations must be even'
            raise InvalidCubeStateError(msg)

    def check_edge_colors(self, ep: list[int], eo: list[int]) -> None:
        for i, (edge_pos, _edge_ori) in enumerate(zip(ep, eo, strict=True)):
            edge_facelets = [
                self._state[facelet]
                for facelet in EDGE_FACELET_MAP[edge_pos]
            ]
            if len(set(edge_facelets)) != 2:
                msg = (
                    f'Edge { i } must have 2 different colors, '
                    f'got { edge_facelets }'
                )
                raise InvalidCubeStateError(msg)

            color1, color2 = edge_facelets
            if color1 in OPPOSITE_FACES and OPPOSITE_FACES[color1] == color2:
                msg = (
                    f'Edge { i } cannot have opposite colors '
                    f'{ color1 } and { color2 }'
                )
                raise InvalidCubeStateError(msg)

    def check_permutation_parity(self, cp: list[int], ep: list[int]) -> None:
        corner_parity = count_inversions(cp) % 2
        edge_parity = count_inversions(ep) % 2

        if corner_parity != edge_parity:
            msg = 'Corner and edge permutation parities must be equal'
            raise InvalidCubeStateError(msg)

    def check_center_orientations(self, so: list[int]) -> None:
        valid_orientations = set(range(self.face_number))

        if len(so) != self.face_number or any(
                orientation not in valid_orientations
                for orientation in so
        ):
            msg = (
                'Center orientation must be between 0 and 5 '
                'for each center'
            )
            raise InvalidCubeStateError(msg)

    def check_face_orientations(self, faces: str) -> tuple[str, str]:
        if not faces:
            msg = 'Specify at leat one face to orient'
            raise InvalidFaceError(msg)

        if len(faces) > 2:
            msg = f'Too much faces ({ len(faces) })'
            raise InvalidFaceError(msg)

        top_face = faces[0]
        front_face = (len(faces) > 1 and faces[1]) or ''

        if top_face not in OPPOSITE_FACES:
            msg = f'{ top_face } is an invalid face'
            raise InvalidFaceError(msg)

        if OPPOSITE_FACES[top_face] == front_face:
            msg = f'{ top_face } { front_face } are opposed faces'
            raise InvalidFaceError(msg)

        if front_face and front_face not in OPPOSITE_FACES:
            msg = f'{ front_face } is an invalid face'
            raise InvalidFaceError(msg)

        return top_face, front_face



import unittest

from cubing_algs.move import Move
from cubing_algs.parsing import parse_moves
from cubing_algs.transform.rotation import compress_final_rotations
from cubing_algs.transform.rotation import compress_rotations
from cubing_algs.transform.rotation import optimize_conjugate_rotations
from cubing_algs.transform.rotation import optimize_double_rotations
from cubing_algs.transform.rotation import optimize_triple_rotations
from cubing_algs.transform.rotation import remove_final_rotations
from cubing_algs.transform.rotation import split_moves_final_rotations


class TransformRemoveFinalRotationsTestCase(unittest.TestCase):

    def test_remove_final_rotations(self):
        provide = parse_moves('R2 F U x y2')
        expect = parse_moves('R2 F U')

        result = remove_final_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_remove_final_rotations_timed(self):
        provide = parse_moves('R2@1 F@2 U@3 x@4 y2@5')
        expect = parse_moves('R2@1 F@2 U@3')

        result = remove_final_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_remove_final_rotations_timed_paused(self):
        provide = parse_moves('R2@1 F@2 U@3 x@4 .@5 y2@6')
        expect = parse_moves('R2@1 F@2 U@3')

        result = remove_final_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

        provide = parse_moves('R2@1 F@2 U@3 .@4 x@5 .@6 y2@7')
        expect = parse_moves('R2@1 F@2 U@3')

        result = remove_final_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))


class SplitMovesFinalRotationsTestCase(unittest.TestCase):

    def test_split_moves_final_rotations(self):
        provide = parse_moves("R2 F x x x'")
        expect = (parse_moves('R2 F'), parse_moves("x x x'"))

        result = split_moves_final_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result[1]:
            self.assertTrue(isinstance(m, Move))

    def test_split_moves_final_rotations_empty(self):
        provide = parse_moves('R2 F')
        expect = (parse_moves('R2 F'), [])

        result = split_moves_final_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

    def test_split_moves_final_rotations_start(self):
        provide = parse_moves('x R2 F')
        expect = (parse_moves('x R2 F'), [])

        result = split_moves_final_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )


class TransformOptimizeTripleRotationsTestCase(unittest.TestCase):

    def test_optimize_triple_rotations(self):
        provide = parse_moves('x2 y2 z2')
        expect = []

        result = optimize_triple_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        provide = parse_moves('y2 x2 z2')

        result = optimize_triple_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        provide = parse_moves('z2 x2 y2')

        result = optimize_triple_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

    def test_optimize_triple_rotations_timed(self):
        provide = parse_moves('x2@0 y2@50 z2@100')
        expect = []

        result = optimize_triple_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

    def test_optimize_triple_rotations_start(self):
        provide = parse_moves('x2 x2 y2 z2')
        expect = parse_moves('x2')

        result = optimize_triple_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_optimize_triple_rotations_end(self):
        provide = parse_moves('x2 y2 z2 x2')
        expect = parse_moves('x2')

        result = optimize_triple_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_optimize_triple_rotations_max(self):
        provide = parse_moves('x2 y2 z2')

        result = optimize_triple_rotations(provide, 0)

        self.assertEqual(
            result,
            provide,
        )


class TransformOptimizeDoubleRotationsTestCase(unittest.TestCase):

    def test_optimize_double_rotations(self):
        provide = parse_moves('x2 y2')
        expect = parse_moves('z2')

        result = optimize_double_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

        provide = parse_moves('z2 x2')
        expect = parse_moves('y2')

        result = optimize_double_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

        provide = parse_moves('z2 y2')
        expect = parse_moves('x2')

        result = optimize_double_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_optimize_double_rotations_timed(self):
        provide = parse_moves('x2@50 y2@100')
        expect = parse_moves('z2')

        result = optimize_double_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_optimize_double_rotations_start(self):
        provide = parse_moves('x x2 y2')
        expect = parse_moves('x z2')

        result = optimize_double_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_optimize_double_rotations_end(self):
        provide = parse_moves('x2 y2 x')
        expect = parse_moves('z2 x')

        result = optimize_double_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_optimize_double_rotations_multiple(self):
        provide = parse_moves('x2 y2 x2')
        expect = parse_moves('y2')

        result = optimize_double_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_optimize_double_rotations_max(self):
        provide = parse_moves('x2 y2')

        result = optimize_double_rotations(provide, 0)

        self.assertEqual(
            result,
            provide,
        )


class TransformOptimizeConjugateRotationsTestCase(unittest.TestCase):

    def test_optimize_conjugate_rotations(self):
        provide = parse_moves("y x2 y'")
        expect = parse_moves('z2')

        result = optimize_conjugate_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

        provide = parse_moves("y' x2 y")
        expect = parse_moves('z2')

        result = optimize_conjugate_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

        provide = parse_moves("z x2 z'")
        expect = parse_moves('y2')

        result = optimize_conjugate_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

        provide = parse_moves("z' y2 z")
        expect = parse_moves('x2')

        result = optimize_conjugate_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_optimize_conjugate_rotations_timed(self):
        provide = parse_moves("y@0 x2@50 y'@100")
        expect = parse_moves('z2')

        result = optimize_conjugate_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_optimize_conjugate_rotations_start(self):
        provide = parse_moves("x x y2 x'")
        expect = parse_moves('x z2')

        result = optimize_conjugate_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_optimize_conjugate_rotations_end(self):
        provide = parse_moves("x y2 x' x")
        expect = parse_moves('z2 x')

        result = optimize_conjugate_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_optimize_conjugate_rotations_multiple(self):
        provide = parse_moves("x' z2 x y x2 y'")
        expect = parse_moves('y2 z2')

        result = optimize_conjugate_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_optimize_conjugate_rotations_max(self):
        provide = parse_moves("y x2 y'")

        result = optimize_conjugate_rotations(provide, 0)

        self.assertEqual(
            result,
            provide,
        )


class TransformCompressRotationsTestCase(unittest.TestCase):

    def test_compress_rotations(self):
        provide = parse_moves("x' z2 x y x2 y'")
        expect = parse_moves('x2')

        result = compress_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_compress_rotations_issues_01(self):
        provide = parse_moves("z@27089 y y z' z' z x x")
        expect = []

        result = compress_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_compress_rotations_impossible(self):
        provide = parse_moves('x')

        result = compress_rotations(provide)

        self.assertEqual(
            result,
            provide,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_compress_rotations_max(self):
        provide = parse_moves("x' z2 x y x2 y'")

        result = compress_rotations(provide, 0)

        self.assertEqual(
            result,
            provide,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))


class TransformCompressFinalRotationsTestCase(unittest.TestCase):

    def test_compress_final_rotations(self):
        provide = parse_moves("R2 F x x x' x x x")
        expect = parse_moves('R2 F')

        result = compress_final_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_compress_final_rotations_empty(self):
        provide = parse_moves('R2 F')
        expect = parse_moves('R2 F')

        result = compress_final_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_compress_final_rotations_timed(self):
        provide = parse_moves("R2@1 F@2 x'@3 x@4")
        expect = parse_moves('R2@1 F@2')

        result = compress_final_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_compress_final_rotations_impair(self):
        provide = parse_moves("R2 F x' x x'")
        expect = parse_moves("R2 F x'")

        result = compress_final_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_compress_final_rotations_double_double(self):
        provide = parse_moves('R2 F x2 z2')
        expect = parse_moves('R2 F y2')

        result = compress_final_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

        provide = parse_moves('R2 F x2 y2')
        expect = parse_moves('R2 F z2')

        result = compress_final_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

        provide = parse_moves('R2 F y2 z2')
        expect = parse_moves('R2 F x2')

        result = compress_final_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_compress_final_rotations_trible_double(self):
        provide = parse_moves('R2 F x2 z2 y2')
        expect = parse_moves('R2 F')

        result = compress_final_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_compress_final_rotations_trible_double_failed(self):
        provide = parse_moves('R2 F x2 z2 y')
        expect = parse_moves("R2 F y'")

        result = compress_final_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_compress_final_rotations_simple_double_simple(self):
        provide = parse_moves("R2 F x z2 x'")
        expect = parse_moves('R2 F y2')

        result = compress_final_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

        provide = parse_moves("R2 F x' z2 x")
        expect = parse_moves('R2 F y2')

        result = compress_final_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_compress_final_rotations_simple_double_simple_clear(self):
        provide = parse_moves("R2 F x z2 x' y2")
        expect = parse_moves('R2 F')

        result = compress_final_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

        provide = parse_moves("R2 F x' z2 x y2")
        expect = parse_moves('R2 F')

        result = compress_final_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_compress_final_rotations_complex(self):
        provide = parse_moves("R2 F z2 x z2 x' y2 x x y2")
        expect = parse_moves('R2 F')

        result = compress_final_rotations(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))



from cubing_algs.constants import MAX_ITERATIONS
from cubing_algs.move import Move
from cubing_algs.transform.optimize import optimize_do_undo_moves
from cubing_algs.transform.optimize import optimize_double_moves
from cubing_algs.transform.optimize import optimize_repeat_three_moves
from cubing_algs.transform.optimize import optimize_triple_moves


def compress_moves(
        old_moves: list[Move],
        max_iterations: int = MAX_ITERATIONS,
) -> list[Move]:
    moves = old_moves.copy()

    for _ in range(max_iterations):
        start_length = len(moves)

        for optimizer in (
            optimize_do_undo_moves,
            optimize_repeat_three_moves,
            optimize_double_moves,
            optimize_triple_moves,
        ):
            moves = optimizer(moves)

        if len(moves) == start_length:
            break

    return moves


def expand_moves(old_moves: list[Move]) -> list[Move]:
    moves: list[Move] = []

    for move in old_moves:
        if move.is_double:
            moves.extend((move.doubled, move.doubled))
        else:
            moves.append(move)

    return moves



from collections.abc import Callable

from cubing_algs.constants import PAUSE_CHAR
from cubing_algs.move import Move


def unpause_moves(old_moves: list[Move]) -> list[Move]:
    moves = []
    for move in old_moves:
        if not move.is_pause:
            moves.append(move)

    return moves


def pause_moves(
        speed: int = 200, factor: int = 2,
        *, multiple: bool = False,
) -> Callable[[list[Move]], list[Move]]:
    """
    Create a configurable pause_moves function.

    Args:
        speed: Base speed in milliseconds (default: 200)
        factor: Multiplier for threshold calculation (default: 2)

    Returns:
        A function that can be used with transform() or called directly.
    """
    def _pause_moves(old_moves: list[Move]) -> list[Move]:
        if not old_moves:
            return old_moves

        if any(not m.is_timed for m in old_moves):
            return old_moves

        moves = []
        threshold = speed * factor

        previous_time = old_moves[0].timed
        for move in old_moves:
            time = move.timed
            delta = time - previous_time

            if delta > threshold:
                if multiple:
                    delta = time - previous_time
                    occurences = int(delta / threshold)
                    offset = (delta - (threshold * occurences)) / 2

                    for i in range(occurences):
                        new_time = int(
                            previous_time + offset + (
                                (i + 1) * threshold
                            ),
                        )
                        moves.append(
                            Move(
                                f'{ PAUSE_CHAR }@{ new_time }',
                            ),
                        )
                    moves.append(move)
                else:
                    offset = int(delta / 2)

                    moves.extend(
                        [
                            Move(f'{ PAUSE_CHAR }@{ time - offset }'),
                            move,
                        ],
                    )
            else:
                moves.append(move)
            previous_time = time

        return moves

    return _pause_moves



from cubing_algs.constants import SYMMETRY_TABLE
from cubing_algs.constants import WIDE_CHAR
from cubing_algs.move import Move


def symmetry_moves(
        old_moves: list[Move],
        ignore_moves: set[str],
        symmetry_table: dict[str, str],
) -> list[Move]:
    moves = []

    for move in old_moves:
        if move.is_pause or move.base_move in ignore_moves:
            moves.append(move)
        else:
            symmetry_move = symmetry_table[move.base_move]

            if move.is_wide_move:
                symmetry_move += WIDE_CHAR

            new_move = Move(
                move.layer + symmetry_move + move.time,
            )

            if move.is_sign_move:
                new_move = new_move.to_sign

            if move.is_double:
                moves.append(new_move.doubled)
            elif move.is_clockwise:
                moves.append(new_move.inverted)
            else:
                moves.append(new_move)

    return moves


def symmetry_type_moves(
        old_moves: list[Move],
        symmetry_type: str,
) -> list[Move]:
    ignore_moves, symmetry_table = SYMMETRY_TABLE[symmetry_type]
    return symmetry_moves(old_moves, ignore_moves, symmetry_table)


def symmetry_m_moves(old_moves: list[Move]) -> list[Move]:
    return symmetry_type_moves(old_moves, 'M')


def symmetry_s_moves(old_moves: list[Move]) -> list[Move]:
    return symmetry_type_moves(old_moves, 'S')


def symmetry_e_moves(old_moves: list[Move]) -> list[Move]:
    return symmetry_type_moves(old_moves, 'E')


def symmetry_c_moves(old_moves: list[Move]) -> list[Move]:
    moves = symmetry_m_moves(old_moves)
    return symmetry_s_moves(moves)



from cubing_algs.constants import OFFSET_TABLE
from cubing_algs.constants import WIDE_CHAR
from cubing_algs.move import Move


def rotate(old_moves: list[Move], rotation: str) -> list[Move]:
    moves: list[Move] = []
    rotation_table: dict[str, str] = OFFSET_TABLE[rotation]

    for move in old_moves:
        layer = move.layer
        time = move.time
        base_move = move.base_move
        wide = WIDE_CHAR if move.is_wide_move else ''

        new_move = move

        if base_move in rotation_table:
            new_move = Move(
                layer + rotation_table[base_move] + wide + time,
            )
            if move.is_counter_clockwise:
                new_move = new_move.inverted
            elif move.is_double:
                new_move = new_move.doubled

            if move.is_sign_move:
                new_move = new_move.to_sign

        moves.append(new_move)

    return moves


def offset_moves(
        old_moves: list[Move],
        rotation: str,
        count: int = 1,
) -> list[Move]:
    result = old_moves
    for _ in range(count):
        result = rotate(result, rotation)
    return result


def offset_x_moves(old_moves: list[Move]) -> list[Move]:
    return offset_moves(old_moves, "x'")


def offset_x2_moves(old_moves: list[Move]) -> list[Move]:
    return offset_moves(old_moves, 'x', 2)


def offset_xprime_moves(old_moves: list[Move]) -> list[Move]:
    return offset_moves(old_moves, 'x')


def offset_y_moves(old_moves: list[Move]) -> list[Move]:
    return offset_moves(old_moves, "y'")


def offset_y2_moves(old_moves: list[Move]) -> list[Move]:
    return offset_moves(old_moves, 'y', 2)


def offset_yprime_moves(old_moves: list[Move]) -> list[Move]:
    return offset_moves(old_moves, 'y')


def offset_z_moves(old_moves: list[Move]) -> list[Move]:
    return offset_moves(old_moves, "z'")


def offset_z2_moves(old_moves: list[Move]) -> list[Move]:
    return offset_moves(old_moves, 'z', 2)


def offset_zprime_moves(old_moves: list[Move]) -> list[Move]:
    return offset_moves(old_moves, 'z')



class InvalidFaceError(Exception):
    """
    Exception raised when an invalid face is encountered.
    """


class InvalidCubeStateError(Exception):
    """
    Exception raised when an invalid cube is encountered.
    """


class InvalidMoveError(Exception):
    """
    Exception raised when an invalid move notation is encountered.

    This can occur when parsing algorithms with incorrect or unsupported
    move notations.
    """


class InvalidBracketError(InvalidMoveError):
    """
    Exception raised when an invalid bracket formation is encountered.
    """


class InvalidOperatorError(InvalidMoveError):
    """
    Exception raised when an invalid bracket formation is encountered.
    """



from cubing_algs.algorithm import Algorithm
from cubing_algs.constants import FACE_INDEXES
from cubing_algs.constants import FACE_ORDER
from cubing_algs.constants import FRONT_FACE_TRANSLATIONS
from cubing_algs.constants import INITIAL_STATE
from cubing_algs.constants import TOP_FACE_TRANSLATIONS
from cubing_algs.display import VCubeDisplay
from cubing_algs.exceptions import InvalidMoveError
from cubing_algs.extensions import rotate  # type: ignore[attr-defined]
from cubing_algs.facelets import cubies_to_facelets
from cubing_algs.facelets import facelets_to_cubies
from cubing_algs.integrity import VCubeIntegrityChecker


class VCube(VCubeIntegrityChecker):
    """
    Virtual 3x3 cube for tracking moves on facelets
    """
    size = 3
    face_number = 6
    face_size = size * size

    def __init__(self, initial: str | None = None, *,
                 check: bool = True,
                 history: list[str] | None = None):
        if initial:
            self._state = initial
            if check:
                self.check_integrity()
        else:
            self._state = INITIAL_STATE

        self.history: list[str] = history or []

    @property
    def state(self) -> str:
        return self._state

    @staticmethod
    def from_cubies(cp: list[int], co: list[int],
                    ep: list[int], eo: list[int],
                    so: list[int]) -> 'VCube':
        return VCube(cubies_to_facelets(cp, co, ep, eo, so))

    @property
    def to_cubies(self) -> tuple[
            list[int], list[int], list[int], list[int], list[int],
    ]:
        return facelets_to_cubies(self._state)

    @property
    def is_solved(self) -> bool:
        return all(face * self.face_size in self.state for face in FACE_ORDER)

    def rotate(self, moves: str | Algorithm) -> str:
        if isinstance(moves, Algorithm):
            for m in moves:
                self.rotate_move(str(m))
        else:
            for m in moves.split(' '):
                self.rotate_move(m)
        return self._state

    def rotate_move(self, move: str, *, history: bool = True) -> str:
        try:
            self._state = rotate.rotate_move(self._state, move)
        except ValueError as e:
            raise InvalidMoveError(str(e)) from e
        else:
            if history:
                self.history.append(move)
            return self._state

    def copy(self, *, full: bool = False) -> 'VCube':
        history = None
        if full:
            history = list(self.history)

        return VCube(
            self.state,
            check=False,
            history=history,
        )

    def oriented_copy(self, faces: str, *, full: bool = False) -> 'VCube':
        top_face, front_face = self.check_face_orientations(faces)
        cube = self.copy(full=full)

        top_face_index = cube.get_face_index(top_face)
        if top_face_index:
            top_rotation = TOP_FACE_TRANSLATIONS[top_face_index]
            cube.rotate_move(top_rotation, history=full)

        if front_face:
            front_face_index = cube.get_face_index(front_face)
            delta = front_face_index - 2  # F index

            if delta:
                front_rotation = FRONT_FACE_TRANSLATIONS[delta]
                cube.rotate_move(front_rotation, history=full)

        return cube

    def display(self, mode: str = '', orientation: str = '') -> str:
        return VCubeDisplay(self).display(mode, orientation)

    def show(self, mode: str = '', orientation: str = '') -> None:
        print(self.display(mode, orientation), end='')

    def get_face(self, face: str) -> str:
        index = FACE_INDEXES[face]
        return self._state[index * self.face_size: (index + 1) * self.face_size]

    def get_face_center_indexes(self) -> list[str]:
        face_centers = []

        for i in range(6):
            face_centers.append(self.state[(i * self.face_size) + 4])

        return face_centers

    def get_face_index(self, face: str) -> int:
        return self.get_face_center_indexes().index(face)

    def get_face_by_center(self, face: str) -> str:
        index = self.get_face_index(face)

        return self._state[index * self.face_size: (index + 1) * self.face_size]

    def __str__(self) -> str:
        """
        Return the facelets of the cube
        """
        faces = []
        for face in FACE_ORDER:
            faces.append(f'{ face }: { self.get_face(face)}')

        return '\n'.join(faces)

    def __repr__(self) -> str:
        """
        Return a string representation that can be used
        to recreate the VCube.
        """
        return f"VCube('{ self._state }')"



import unittest

from cubing_algs.move import Move
from cubing_algs.parsing import parse_moves
from cubing_algs.transform.offset import offset_x2_moves
from cubing_algs.transform.offset import offset_x_moves
from cubing_algs.transform.offset import offset_xprime_moves
from cubing_algs.transform.offset import offset_y2_moves
from cubing_algs.transform.offset import offset_y_moves
from cubing_algs.transform.offset import offset_yprime_moves
from cubing_algs.transform.offset import offset_z2_moves
from cubing_algs.transform.offset import offset_z_moves
from cubing_algs.transform.offset import offset_zprime_moves


class TransformOffsetTestCase(unittest.TestCase):

    def test_offset_x_moves(self):
        provide = parse_moves("R U R' U'")
        expect = parse_moves("R B R' B'")

        result = offset_x_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_offset_x_moves_wide_standard(self):
        provide = parse_moves("R U Rw' Uw'")
        expect = parse_moves("R B Rw' Bw'")

        result = offset_x_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_offset_x_moves_wide_sign(self):
        provide = parse_moves("R U r' u'")
        expect = parse_moves("R B r' b'")

        result = offset_x_moves(provide)

        self.assertEqual(
            result,
            expect,
        )

        for m in result:
            self.assertTrue(isinstance(m, Move))

    def test_offset_x2_moves(self):
        provide = parse_moves("R U R' U'")
        expect = parse_moves("R D R' D'")

        self.assertEqual(
            offset_x2_moves(provide),
            expect,
        )

    def test_offset_xprime_moves(self):
        provide = parse_moves("R U R' U'")
        expect = parse_moves("R F R' F'")

        self.assertEqual(
            offset_xprime_moves(provide),
            expect,
        )

    def test_offset_y_moves(self):
        provide = parse_moves("R U R' U'")
        expect = parse_moves("F U F' U'")

        self.assertEqual(
            offset_y_moves(provide),
            expect,
        )

    def test_offset_y2_moves(self):
        provide = parse_moves("R U R' U'")
        expect = parse_moves("L U L' U'")

        self.assertEqual(
            offset_y2_moves(provide),
            expect,
        )

    def test_offset_yprime_moves(self):
        provide = parse_moves("R U R' U'")
        expect = parse_moves("B U B' U'")

        self.assertEqual(
            offset_yprime_moves(provide),
            expect,
        )

    def test_offset_z_moves(self):
        provide = parse_moves("R U R' U'")
        expect = parse_moves("D R D' R'")

        self.assertEqual(
            offset_z_moves(provide),
            expect,
        )

    def test_offset_z2_moves(self):
        provide = parse_moves("R U R' U'")
        expect = parse_moves("L D L' D'")

        self.assertEqual(
            offset_z2_moves(provide),
            expect,
        )

    def test_offset_zprime_moves(self):
        provide = parse_moves("R U R' U'")
        expect = parse_moves("U L U' L'")

        self.assertEqual(
            offset_zprime_moves(provide),
            expect,
        )

    def test_offset_big_moves(self):
        provide = parse_moves('3R')
        expect = parse_moves('3L')

        self.assertEqual(
            offset_y2_moves(provide),
            expect,
        )

        provide = parse_moves('3R2')
        expect = parse_moves('3L2')

        self.assertEqual(
            offset_y2_moves(provide),
            expect,
        )

        provide = parse_moves("3R'")
        expect = parse_moves("3L'")

        self.assertEqual(
            offset_y2_moves(provide),
            expect,
        )

    def test_offset_big_moves_timed(self):
        provide = parse_moves('3R@100')
        expect = parse_moves('3L@100')

        self.assertEqual(
            offset_y2_moves(provide),
            expect,
        )

        provide = parse_moves('3R2@100')
        expect = parse_moves('3L2@100')

        self.assertEqual(
            offset_y2_moves(provide),
            expect,
        )

        provide = parse_moves("3R'@100")
        expect = parse_moves("3L'@100")

        self.assertEqual(
            offset_y2_moves(provide),
            expect,
        )

    def test_offset_big_moves_timed_with_pauses(self):
        provide = parse_moves('.@50 3R@100 .@150')
        expect = parse_moves('.@50 3L@100 .@150')

        self.assertEqual(
            offset_y2_moves(provide),
            expect,
        )



# ruff: noqa: E241 E501
import unittest

from cubing_algs.constants import INNER_MOVES
from cubing_algs.constants import ROTATIONS
from cubing_algs.move import Move
from cubing_algs.parsing import split_moves
from cubing_algs.transform.degrip import degrip_full_moves
from cubing_algs.transform.degrip import degrip_x_moves
from cubing_algs.transform.degrip import degrip_y_moves
from cubing_algs.transform.degrip import degrip_z_moves
from cubing_algs.transform.rotation import remove_final_rotations
from cubing_algs.transform.size import compress_moves


class TransformDegripTestCase(unittest.TestCase):
    maxDiff = None

    def check_basic_grip_degrip(self, provide, expect, dimension,
                                function, invert):
        if invert:
            provide = split_moves(f"{ dimension }'{ provide }{ dimension }")
        else:
            provide = split_moves(f"{ dimension }{ provide }{ dimension }'")

        expect = [expect]

        with self.subTest(provide=provide, expect=expect):
            degripped = remove_final_rotations(
                function(provide),
            )

            self.assertEqual(
                degripped,
                remove_final_rotations(
                    expect,
                ),
            )

    def check_degrip(self, provide, expect, function, name):
        provide = split_moves(provide)
        expect = split_moves(expect)

        with self.subTest(name=name, provide=provide, expect=expect):

            degripped = compress_moves(
                remove_final_rotations(
                    function(provide),
                ),
            )

            self.assertEqual(
                degripped,
                compress_moves(
                    remove_final_rotations(
                        expect,
                    ),
                ),
            )
            for m in degripped:
                self.assertTrue(isinstance(m, Move))

    def test_basic_x_grip_degrip(self):
        for _move, _expected in zip(
                ('R', 'F', 'U', 'L', 'B', 'D', 'M', 'S', 'E', 'x', 'y', 'z'),
                ('R', 'D', 'F', 'L', 'U', 'B', 'M', 'E', "S'", 'x', 'z', "y'"),
                strict=True,
        ):
            move = Move(_move)
            expected = Move(_expected)

            self.check_basic_grip_degrip(
                move, expected,
                'x', degrip_x_moves,
                invert=False,
            )
            self.check_basic_grip_degrip(
                move.inverted, expected.inverted,
                'x', degrip_x_moves,
                invert=False,
            )

            if move not in ROTATIONS + INNER_MOVES:
                self.check_basic_grip_degrip(
                    move.lower(), expected.lower(),
                    'x', degrip_x_moves,
                    invert=False,
                )

        for _move, _expected in zip(
                ('R', 'F', 'U', 'L', 'B', 'D', 'M', 'S', 'E', 'x', 'y', 'z'),
                ('R', 'U', 'B', 'L', 'D', 'F', 'M', "E'", 'S', 'x', "z'", 'y'),
                strict=True,
        ):
            move = Move(_move)
            expected = Move(_expected)

            self.check_basic_grip_degrip(
                move, expected,
                'x', degrip_x_moves,
                invert=True,
            )
            self.check_basic_grip_degrip(
                move.inverted, expected.inverted,
                'x', degrip_x_moves,
                invert=True,
            )

            if move not in ROTATIONS + INNER_MOVES:
                self.check_basic_grip_degrip(
                    move.lower(), expected.lower(),
                    'x', degrip_x_moves,
                    invert=True,
                )

    def test_basic_y_grip_degrip(self):
        for _move, _expected in zip(
                ('R', 'F', 'U', 'L', 'B', 'D', 'M', 'S', 'E', 'x', 'y', 'z'),
                ('B', 'R', 'U', 'F', 'L', 'D', 'S', "M'", 'E', "z'", 'y', 'x'),
                strict=True,
        ):
            move = Move(_move)
            expected = Move(_expected)

            self.check_basic_grip_degrip(
                move, expected,
                'y', degrip_y_moves,
                invert=False,
            )
            self.check_basic_grip_degrip(
                move.inverted, expected.inverted,
                'y', degrip_y_moves,
                invert=False,
            )
            if move not in ROTATIONS + INNER_MOVES:
                self.check_basic_grip_degrip(
                    move.lower(), expected.lower(),
                    'y', degrip_y_moves,
                    invert=False,
                )

        for _move, _expected in zip(
                ('R', 'F', 'U', 'L', 'B', 'D', 'M', 'S', 'E', 'x', 'y', 'z'),
                ('F', 'L', 'U', 'B', 'R', 'D', "S'", 'M', 'E', 'z', 'y', "x'"),
                strict=True,
        ):
            move = Move(_move)
            expected = Move(_expected)

            self.check_basic_grip_degrip(
                move, expected,
                'y', degrip_y_moves,
                invert=True,
            )
            self.check_basic_grip_degrip(
                move.inverted, expected.inverted,
                'y', degrip_y_moves,
                invert=True,
            )
            if move not in ROTATIONS + INNER_MOVES:
                self.check_basic_grip_degrip(
                    move.lower(), expected.lower(),
                    'y', degrip_y_moves,
                    invert=True,
                )

    def test_basic_z_grip_degrip(self):
        for _move, _expected in zip(
                ('R', 'F', 'U', 'L', 'B', 'D', 'M', 'S', 'E', 'x', 'y', 'z'),
                ('U', 'F', 'L', 'D', 'B', 'R', 'E', 'S', "M'", 'y', "x'", 'z'),
                strict=True,
        ):
            move = Move(_move)
            expected = Move(_expected)

            self.check_basic_grip_degrip(
                move, expected,
                'z', degrip_z_moves,
                invert=False,
            )
            self.check_basic_grip_degrip(
                move.inverted, expected.inverted,
                'z', degrip_z_moves,
                invert=False,
            )
            if move not in ROTATIONS + INNER_MOVES:
                self.check_basic_grip_degrip(
                    move.lower(), expected.lower(),
                    'z', degrip_z_moves,
                    invert=False,
                )

        for _move, _expected in zip(
                ('R', 'F', 'U', 'L', 'B', 'D', 'M', 'S', 'E', 'x', 'y', 'z'),
                ('D', 'F', 'R', 'U', 'B', 'L', "E'", 'S', 'M', "y'", 'x', 'z'),
                strict=True,
        ):
            move = Move(_move)
            expected = Move(_expected)

            self.check_basic_grip_degrip(
                move, expected,
                'z', degrip_z_moves,
                invert=True,
            )
            self.check_basic_grip_degrip(
                move.inverted, expected.inverted,
                'z', degrip_z_moves,
                invert=True,
            )
            if move not in ROTATIONS + INNER_MOVES:
                self.check_basic_grip_degrip(
                    move.lower(), expected.lower(),
                    'z', degrip_z_moves,
                    invert=True,
                )

    def test_start_degrip_x(self):
        basic_moves = 'RF'
        provide = 'x' + basic_moves
        expect = 'RD'

        self.check_degrip(
            provide, expect,
            degrip_x_moves,
            'Start Degrip X',
        )

    def test_end_degrip_x(self):
        basic_moves = 'RF'
        provide = basic_moves + 'x'
        expect = 'RF'

        self.check_degrip(
            provide, expect,
            degrip_x_moves,
            'End Degrip X',
        )

    def test_middle_degrip_x(self):
        provide = 'R' + 'x' + 'F'
        expect = 'RD'

        self.check_degrip(
            provide, expect,
            degrip_x_moves,
            'Middle Degrip X',
        )

    def test_degrip_x(self):
        base_algo = "RUR'U'"

        for prefix, suffix, expect, name in [
                ('', '',               "RUR'U'",             'No Degrip'),
                ('x', '',              "RFR'F'x",            'Start'),
                ('', "x'",             "RUR'U'x'",           'End'),
                ('x', "x'",            "RFR'F'",             'Start/End'),
                ("x'", 'x',            "RBR'B'",             'Start/End/Revert'),
                ('Rx', '',             "R2FR'F'x",           'Prefix'),
                ('RUFLDBx', '',        "RUFLDBRFR'F'x",      'Prefix Bis'),
                ("RUFLDBx'", '',       "RUFLDBRBR'B'x'",     'Prefix Bis Inverted'),
                ('', "x'R",            "RUR'U'Rx'",          'Suffix'),
                ('', "x'RUFLDB",       "RUR'U'RBULFDx'",     'Suffix Bis'),
                ('', 'xRUFLDB',        "RUR'U'RFDLBUx",      'Suffix Bis Inverted'),
                ('Fx', "x'F'",         "FRFR'F2",            'Enclosed'),
                ('FRUx', "x'F'R'U'",   "FRURFR'F2R'U'",      'Enclosed Bis'),
                ("FRUx'", "xF'R'U'",   "FRURBR'B'F'R'U'",    'Enclosed Bis Inverted'),
                ("FRUx'", "x'F'R'U'",  "FRURBR'B2R'D'x2",    'Enclosed Bis Double'),
                ('Rx', "x'FRUxFRU",    'R2FUDRFx',           'Triple'),
                ("Rx'", "x'FRUx'FRU",  'R2BD2RFx',           'Triple Bis'),
        ]: # noqa E241
            algo = f'{ prefix }{ base_algo }{ suffix }'
            self.check_degrip(
                algo,
                expect,
                degrip_x_moves,
                name,
            )

    def test_degrip_y(self):
        base_algo = "RUR'U'"

        for prefix, suffix, expect, name in [
                ('', '',               "RUR'U'",             'No Degrip'),
                ('y', '',              "BUB'U'y",            'Start'),
                ('', "y'",             "RUR'U'y'",           'End'),
                ('y', "y'",            "BUB'U'",             'Start/End'),
                ("y'", 'y',            "FUF'U'",             'Start/End/Revert'),
                ('Ry', '',             "RBUB'U'y",           'Prefix'),
                ('RUFLDBy', '',        "RUFLDB2UB'U'y",      'Prefix Bis'),
                ("RUFLDBy'", '',       "RUFLDBFUF'U'y'",     'Prefix Bis Inverted'),
                ('', "y'R",            "RUR'U'Fy'",          'Suffix'),
                ('', "y'RUFLDB",       "RUR'U'FULBDRy'",     'Suffix Bis'),
                ('', 'yRUFLDB',        "RUR'U'BURFDLy",      'Suffix Bis Inverted'),
                ('Fy', "y'F'",         "FBUB'U'F'",          'Enclosed'),
                ('FRUy', "y'F'R'U'",   "FRUBUB'U'F'R'U'",    'Enclosed Bis'),
                ("FRUy'", "yF'R'U'",   "FRUFUF'U'F'R'U'",    'Enclosed Bis Inverted'),
                ("FRUy'", "y'F'R'U'",  "FRUFUF'U'B'L'U'y2",  'Enclosed Bis Double'),
                ('Ry', "y'FRUyFRU",    "RBUB'U'FRURBUy",     'Triple'),
        ]:
            algo = f'{ prefix }{ base_algo }{ suffix }'
            self.check_degrip(
                algo,
                expect,
                degrip_y_moves,
                name,
            )

    def test_degrip_z(self):
        base_algo = "RUR'U'"

        for prefix, suffix, expect, name in [
                ('', '',               "RUR'U'",             'No Degrip'),
                ('z', '',              "ULU'L'z",            'Start'),
                ('', "z'",             "RUR'U'z'",           'End'),
                ('z', "z'",            "ULU'L'",             'Start/End'),
                ("z'", 'z',            "DRD'R'",             'Start/End/Revert'),
                ('Rz', '',             "RULU'L'z",           'Prefix'),
                ('RUFLDBz', '',        "RUFLDBULU'L'z",      'Prefix Bis'),
                ("RUFLDBz'", '',       "RUFLDBDRD'R'z'",     'Prefix Bis Inverted'),
                ('', "z'R",            "RUR'U'Dz'",          'Suffix'),
                ('', "z'RUFLDB",       "RUR'U'DRFULBz'",     'Suffix Bis'),
                ('', 'zRUFLDB',        "RUR'LFDRBz",         'Suffix Bis Inverted'),
                ('Fz', "z'F'",         "FULU'L'F'",          'Enclosed'),
                ('FRUz', "z'F'R'U'",   "FRU2LU'L'F'R'U'",    'Enclosed Bis'),
                ("FRUz'", "zF'R'U'",   "FRUDRD'R'F'R'U'",    'Enclosed Bis Inverted'),
                ("FRUz'", "z'F'R'U'",  "FRUDRD'R'F'L'D'z2",  'Enclosed Bis Double'),
                ('Rz', "z'FRUzFRU",    "RULU'L'FRUFULz",     'Triple'),
        ]:
            algo = f'{ prefix }{ base_algo }{ suffix }'
            self.check_degrip(
                algo,
                expect,
                degrip_z_moves,
                name,
            )

    def test_multi_degrip_full_moves(self):
        self.check_degrip(
            ''.join(['y2', 'z2', 'M2', "D'", 'M2', 'D2', 'M2', "D'", 'M2', 'z2']),  # noqa: FLY002
            ''.join(['M2', "U'", 'M2', 'U2', 'M2', "U'", 'M2', 'z2', 'y2', 'z2']),  # noqa: FLY002
            degrip_full_moves,
            'Multi move degrip OK',
        )

        self.check_degrip(
            ''.join(['z2', 'y2', 'z2', 'M2', "U'", 'M2', 'U2', 'M2', "U'", 'M2']),  # noqa: FLY002
            ''.join(['M2', "U'", 'M2', 'U2', 'M2', "U'", 'M2', 'z2', 'y2', 'z2']),  # noqa: FLY002
            degrip_full_moves,
            'Multi move degrip KO',
        )

    def test_degrip_full_moves(self):
        for provide, expect, name in [
                ("RUR'U'", "RUR'U'", 'No Degrip'),
                ("R2U'R'URU'x'U'z'U'RU'R'U'rB", "R2U'R'URU'B'R'FR'F'R'fDz'y'", 'Complex 1'),
                ("x'U'R'UL'U'R2U'R'U2r", "B'R'BL'B'R2B'R'B2rx'", 'Complex 2'),
        ]:
            self.check_degrip(
                provide,
                expect,
                degrip_full_moves,
                name,
            )

    def test_degrip_big_moves(self):
        provide = split_moves('z2 3R')
        expect = split_moves('3L')

        self.assertEqual(
            remove_final_rotations(
                degrip_full_moves(provide),
            ),
            expect,
        )

        provide = split_moves("z2 3R'")
        expect = split_moves("3L'")

        self.assertEqual(
            remove_final_rotations(
                degrip_full_moves(provide),
            ),
            expect,
        )

        provide = split_moves('z2 3R2')
        expect = split_moves('3L2')

        self.assertEqual(
            remove_final_rotations(
                degrip_full_moves(provide),
            ),
            expect,
        )

    def test_degrip_big_moves_timed(self):
        provide = split_moves('z2@100 3R@200')
        expect = split_moves('3L@200')

        self.assertEqual(
            remove_final_rotations(
                degrip_full_moves(provide),
            ),
            expect,
        )

        provide = split_moves("z2@100 3R'@200")
        expect = split_moves("3L'@200")

        self.assertEqual(
            remove_final_rotations(
                degrip_full_moves(provide),
            ),
            expect,
        )

        provide = split_moves('z2@100 3R2@200')
        expect = split_moves('3L2@200')

        self.assertEqual(
            remove_final_rotations(
                degrip_full_moves(provide),
            ),
            expect,
        )

    def test_degrip_big_moves_timed_paused(self):
        provide = split_moves('z2@100 .@150 3R@200')
        expect = split_moves('.@150 3L@200')

        self.assertEqual(
            remove_final_rotations(
                degrip_full_moves(provide),
            ),
            expect,
        )



"""
Utilities for parsing and normalizing Rubik's cube algorithm notations.

This module provides functions to clean, validate, and convert
string representations of Rubik's cube algorithms into structured
Algorithm objects.
"""
import logging
import re

from cubing_algs.algorithm import Algorithm
from cubing_algs.commutator_conjugate import expand_commutators_and_conjugates
from cubing_algs.constants import MOVE_SPLIT
from cubing_algs.exceptions import InvalidMoveError
from cubing_algs.move import Move

logger = logging.getLogger(__name__)

CLEAN_PATTERNS = [
    (re.compile(r'[`’]'), "'"),  # # noqa RUF001
    (re.compile(r'[():,\[\]]'), ' '),
    (re.compile(r'\s+'), ' '),
    (re.compile(r"2'"), '2'),
]

CASE_FIXES = str.maketrans('mseXYZ', 'MSExyz')


def clean_moves(moves: str) -> str:
    """
    Normalize and clean a string representation of moves.

    This function standardizes move notation by:
    - Removing whitespace and unnecessary characters
    - Converting alternative notations to standard ones
    - Standardizing the casing of slice moves
    """
    moves = moves.strip()

    for pattern, replacement in CLEAN_PATTERNS:
        moves = pattern.sub(replacement, moves)

    return moves.translate(CASE_FIXES)


def split_moves(moves: str) -> list[Move]:
    """
    Split a string of moves into individual Move objects.

    Uses the MOVE_SPLIT pattern from constants to identify boundaries
    between individual moves in the string.
    """
    return [
        Move(x.strip())
        for x in MOVE_SPLIT.split(moves)
        if x.strip()
    ]


def check_moves(moves: list[Move]) -> bool:
    """
    Validate a list of Move objects.

    Checks that each move has a valid base move, layer and modifier.
    Logs errors for invalid moves.
    """
    for move in moves:
        if not (move.is_valid_move
                and move.is_valid_modifier
                and move.is_valid_layer):
            if logger.isEnabledFor(logging.ERROR):  # pragma: no cover
                move_string = ''.join(str(m) for m in moves)
                if not move.is_valid_move:
                    logger.error(
                        '"%s" -> %s is not a valid move',
                        move_string, move,
                    )
                elif not move.is_valid_modifier:
                    logger.error(
                        '"%s" -> %s has an invalid modifier',
                        move_string, move,
                    )
                else:
                    logger.error(
                        '"%s" -> %s has an invalid layer',
                        move_string, move,
                    )
            return False

    return True


def parse_moves(raw_moves: str | list[str] | Algorithm,
                secure: bool = True) -> Algorithm:  # noqa: FBT001, FBT002
    """
    Parse raw move data into an Algorithm object.

    This function handles different input types and performs
    cleaning and validation:
    - If raw_moves is already an Algorithm, it's returned as-is
    - If raw_moves is a list, it's joined into a string
    - Strings are cleaned, split into moves, validated, and converted
      to an Algorithm

    Now supports commutators [A, B] and conjugates [A: B].

    Examples:
        [A, B] becomes A B A' B' (commutator)
        [A: B] becomes A B A' (conjugate)
        [[R: U], D] becomes R U R' D R U' R' D'
        [F: [U, R]] becomes F U R U' R' F'
    """
    if isinstance(raw_moves, Algorithm):
        return raw_moves

    if isinstance(raw_moves, list):
        raw_moves = ''.join(str(m) for m in raw_moves)

    expanded_moves = expand_commutators_and_conjugates(raw_moves)

    if not secure:
        moves = split_moves(clean_moves(expanded_moves))
    else:
        moves = split_moves(expanded_moves)

    if not secure and not check_moves(moves):
        error = f'{ raw_moves } contains invalid move'
        raise InvalidMoveError(error)

    return Algorithm(moves)


def parse_moves_cfop(moves: str) -> Algorithm:
    """
    Parse moves specifically for CFOP method algorithms.

    Similar to parse_moves, but also removes typical setup and restoration
    moves (y and U rotations) from the beginning and end of the algorithm.
    This is useful for standardizing CFOP algorithms, which often include
    such moves for convenience.
    """
    algo = parse_moves(moves, secure=False)

    if algo[0].base_move in {'y', 'U'}:
        algo = algo[1:]

    if algo[-1].base_move in {'y', 'U'}:
        algo = algo[:-1]

    return algo



import unittest

from cubing_algs.parsing import parse_moves
from cubing_algs.transform.optimize import optimize_double_moves


class MetricsTestCase(unittest.TestCase):
    maxDiff = None

    def test_metrics(self):
        algo = parse_moves("yM2UMU2M'UM2")
        self.assertEqual(
            algo.metrics,
            {
                'generators': ['M', 'U'],
                'inner_moves': 4,
                'outer_moves': 3,
                'pauses': 0,
                'rotations': 1,
                'htm': 11,
                'qtm': 16,
                'stm': 7,
                'etm': 8,
                'rtm': 1,
                'qstm': 10,
            },
        )

    def test_htm(self):
        moves = ['R', 'R2', 'M', 'M2', 'x2', "f'"]
        scores = [1, 1, 2, 2, 0, 1]

        for move, score in zip(moves, scores, strict=True):
            self.assertEqual(parse_moves(move).metrics['htm'], score)

    def test_qtm(self):
        moves = ['R', 'R2', 'M', 'M2', 'x2', "f'"]
        scores = [1, 2, 2, 4, 0, 1]

        for move, score in zip(moves, scores, strict=True):
            self.assertEqual(parse_moves(move).metrics['qtm'], score)

    def test_stm(self):
        moves = ['R', 'R2', 'M', 'M2', 'x2', "f'"]
        scores = [1, 1, 1, 1, 0, 1]

        for move, score in zip(moves, scores, strict=True):
            self.assertEqual(parse_moves(move).metrics['stm'], score)

    def test_etm(self):
        moves = ['R', 'R2', 'M', 'M2', 'x2', "f'"]
        scores = [1, 1, 1, 1, 1, 1]

        for move, score in zip(moves, scores, strict=True):
            self.assertEqual(parse_moves(move).metrics['etm'], score)

    def test_qstm(self):
        moves = ['R', 'R2', 'M', 'M2', 'x2', "f'"]
        scores = [1, 2, 1, 2, 0, 1]

        for move, score in zip(moves, scores, strict=True):
            self.assertEqual(parse_moves(move).metrics['qstm'], score)

    def test_rtm(self):
        moves = ['R', 'R2', 'M', 'M2', 'x2', "f'", 'x y2']
        scores = [0, 0, 0, 0, 2, 0, 3]

        for move, score in zip(moves, scores, strict=True):
            self.assertEqual(parse_moves(move).metrics['rtm'], score)

    def test_issue_11(self):
        moves = "R U F' B R' U F' U' F D F' D' F' D' F D' L D L' R D' R' D' B D' B' D' D' R D' D' R' D B' D' B D' D' F D' F' D F D F' D' D D' D' L D B D' B' L' D R F D F' D' R' R F D' F' D' F D F' R' F D F' D' F' R F R' D"  # noqa: E501

        algo = parse_moves(moves)
        self.assertEqual(
            algo.metrics,
            {
                'generators': ['D', 'F', 'R', 'B', 'L', 'U'],
                'inner_moves': 0,
                'outer_moves': 80,
                'pauses': 0,
                'rotations': 0,
                'htm': 80,
                'qtm': 80,
                'stm': 80,
                'etm': 80,
                'rtm': 0,
                'qstm': 80,
            },
        )

        compress = algo.transform(optimize_double_moves)

        self.assertEqual(
            compress.metrics,
            {
                'generators': ['D', 'F', 'R', 'B', 'L', 'U'],
                'inner_moves': 0,
                'outer_moves': 76,
                'pauses': 0,
                'rotations': 0,
                'htm': 76,
                'qtm': 80,
                'stm': 76,
                'etm': 76,
                'rtm': 0,
                'qstm': 80,
            },
        )

    def test_metrics_wide_sign(self):
        algo = parse_moves('RFu')
        self.assertEqual(
            algo.metrics,
            {
                'generators': ['R', 'F', 'u'],
                'inner_moves': 0,
                'outer_moves': 3,
                'pauses': 0,
                'rotations': 0,
                'htm': 3,
                'qtm': 3,
                'stm': 3,
                'etm': 3,
                'rtm': 0,
                'qstm': 3,
            },
        )

    def test_metrics_wide_standard(self):
        algo = parse_moves('RFUw')
        self.assertEqual(
            algo.metrics,
            {
                'generators': ['R', 'F', 'Uw'],
                'inner_moves': 0,
                'outer_moves': 3,
                'pauses': 0,
                'rotations': 0,
                'htm': 3,
                'qtm': 3,
                'stm': 3,
                'etm': 3,
                'rtm': 0,
                'qstm': 3,
            },
        )

    def test_metrics_pauses(self):
        algo = parse_moves('R..Fu.')
        self.assertEqual(
            algo.metrics,
            {
                'generators': ['R', 'F', 'u'],
                'inner_moves': 0,
                'outer_moves': 3,
                'pauses': 3,
                'rotations': 0,
                'htm': 3,
                'qtm': 3,
                'stm': 3,
                'etm': 3,
                'rtm': 0,
                'qstm': 3,
            },
        )



from cubing_algs.constants import MAX_ITERATIONS
from cubing_algs.move import Move
from cubing_algs.transform.optimize import optimize_do_undo_moves
from cubing_algs.transform.optimize import optimize_double_moves
from cubing_algs.transform.optimize import optimize_repeat_three_moves
from cubing_algs.transform.optimize import optimize_triple_moves

CANCEL_TRIPLET = {'x2', 'y2', 'z2'}


def remove_final_rotations(old_moves: list[Move]) -> list[Move]:
    moves = []

    rotation = True
    for move in reversed(old_moves):
        if rotation and (move.is_rotation_move or move.is_pause):
            continue
        rotation = False
        moves.append(move)

    return list(reversed(moves))


def optimize_triple_rotations(
        old_moves: list[Move],
        max_depth: int = MAX_ITERATIONS,
) -> list[Move]:
    """
    x2, y2, z2 --> <nothing>
    x2, z2, y2 --> <nothing>
    z2, x2, y2 --> <nothing>
    """
    if max_depth <= 0:
        return old_moves

    i = 0
    changed = False
    moves = old_moves.copy()

    while i < len(moves) - 2:
        triplet = {moves[i].untimed, moves[i + 1].untimed, moves[i + 2].untimed}
        if triplet == CANCEL_TRIPLET:
            moves[i:i + 3] = []
            changed = True
        else:
            i += 1

    if changed:
        return optimize_triple_rotations(moves, max_depth - 1)

    return moves


def optimize_double_rotations(
        old_moves: list[Move],
        max_depth: int = MAX_ITERATIONS,
) -> list[Move]:
    """
    x2, y2 --> z2
    x2, z2 --> y2
    y2, z2 --> x2
    """
    if max_depth <= 0:
        return old_moves

    i = 0
    changed = False
    moves = old_moves.copy()

    while i < len(moves) - 1:
        one = moves[i].untimed
        two = moves[i + 1].untimed
        if (
            one != two
            and one.is_double
            and two.is_double
        ):
            missing_rotation = (CANCEL_TRIPLET - {one, two}).pop()
            moves[i:i + 2] = [Move(missing_rotation)]
            changed = True
        else:
            i += 1

    if changed:
        return optimize_double_moves(moves, max_depth - 1)

    return moves


def optimize_conjugate_rotations(
        old_moves: list[Move],
        max_depth: int = MAX_ITERATIONS,
) -> list[Move]:
    """
    x, y2, x' --> z2
    x', y2, x --> z2
    x, z2, x' --> y2
    y, z2, y' --> x2
    """
    if max_depth <= 0:
        return old_moves

    i = 0
    changed = False
    moves = old_moves.copy()

    while i < len(moves) - 2:
        one = moves[i].untimed
        two = moves[i + 1].untimed
        three = moves[i + 2].untimed
        if (
                two.is_double
                and one.base_move == three.base_move
                and not one.is_double
                and not three.is_double
                and one.modifier != three.modifier
        ):
            missing_rotation = (CANCEL_TRIPLET - {one.doubled, two}).pop()
            moves[i:i + 3] = [Move(missing_rotation)]
            changed = True
        else:
            i += 1

    if changed:
        return optimize_double_moves(moves, max_depth - 1)

    return moves


def split_moves_final_rotations(
        old_moves: list[Move],
) -> tuple[list[Move], list[Move]]:
    moves = []
    rotations = []

    rotation = True
    for move in reversed(old_moves):
        if rotation and (move.is_rotation_move or move.is_pause):
            rotations.append(move)
            continue
        rotation = False
        moves.append(move)

    if not rotations:
        return old_moves, []

    moves.reverse()
    rotations.reverse()

    return moves, rotations


def compress_rotations(
        old_moves: list[Move],
        max_iterations: int = MAX_ITERATIONS,
) -> list[Move]:
    moves = old_moves.copy()

    for _ in range(max_iterations):
        start_length = len(moves)

        for optimizer in (
                optimize_do_undo_moves,
                optimize_repeat_three_moves,
                optimize_double_moves,
                optimize_triple_moves,
                optimize_conjugate_rotations,
                optimize_double_rotations,
                optimize_triple_rotations,
        ):
            moves = optimizer(moves)

        if len(moves) == start_length:
            break

    return moves


def compress_final_rotations(old_moves: list[Move]) -> list[Move]:
    moves, rotations = split_moves_final_rotations(old_moves)

    if len(rotations) > 1:
        rotations = compress_rotations(rotations)

    return moves + rotations



import unittest
from unittest.mock import MagicMock
from unittest.mock import patch

from cubing_algs.commutator_conjugate import expand_commutators_and_conjugates
from cubing_algs.commutator_conjugate import find_innermost_brackets
from cubing_algs.commutator_conjugate import invert_moves
from cubing_algs.commutator_conjugate import split_on_separator
from cubing_algs.exceptions import InvalidBracketError
from cubing_algs.exceptions import InvalidOperatorError


class TestFindInnermostBrackets(unittest.TestCase):

    def test_no_brackets(self):
        """Should return None when no brackets are present"""
        self.assertIsNone(find_innermost_brackets("R U R' U'"))

    def test_single_level_brackets(self):
        """Should find brackets at depth 1"""
        result = find_innermost_brackets("[R U R']")
        self.assertEqual(result, (0, 7))

    def test_nested_brackets(self):
        """Should find the deepest nested brackets"""
        result = find_innermost_brackets('[[R U], D]')
        self.assertEqual(result, (1, 5))  # Inner brackets [R U]

    def test_multiple_nested_brackets(self):
        """Should find first occurrence of deepest brackets"""
        result = find_innermost_brackets('[[R U], [D F]]')
        self.assertEqual(result, (1, 5))  # First inner brackets [R U]

    def test_complex_nesting(self):
        """Should handle complex nested structures"""
        result = find_innermost_brackets('[A [B [C D] E] F]')
        self.assertEqual(result, (6, 10))  # Innermost [C D]

    def test_empty_brackets(self):
        """Should handle empty brackets"""
        result = find_innermost_brackets('[]')
        self.assertEqual(result, (0, 1))

    def test_malformed_brackets_opening_only(self):
        """Should return None for malformed brackets (opening only)"""
        self.assertIsNone(find_innermost_brackets('[R U'))

    def test_malformed_brackets_closing_only(self):
        """Should return None for malformed brackets (closing only)"""
        self.assertIsNone(find_innermost_brackets('R U]'))


class TestSplitOnSeparator(unittest.TestCase):

    def test_no_separator(self):
        """Should return None when separator is not found"""
        self.assertIsNone(split_on_separator("R U R'", ','))

    def test_top_level_comma(self):
        """Should split on comma at top level"""
        result = split_on_separator('R U, D F', ',')
        self.assertEqual(result, ('R U', ' D F'))

    def test_top_level_colon(self):
        """Should split on colon at top level"""
        result = split_on_separator('R U: D F', ':')
        self.assertEqual(result, ('R U', ' D F'))

    def test_separator_inside_brackets(self):
        """Should not split on separator inside brackets"""
        self.assertIsNone(split_on_separator('R [U, D] F', ','))

    def test_nested_brackets_with_separator(self):
        """Should handle nested brackets with separator inside"""
        self.assertIsNone(split_on_separator('[[R, U], D]', ','))

    def test_multiple_separators_top_level(self):
        """Should split on first occurrence at top level"""
        result = split_on_separator('A, B, C', ',')
        self.assertEqual(result, ('A', ' B, C'))

    def test_separator_at_beginning(self):
        """Should handle separator at beginning"""
        result = split_on_separator(',R U', ',')
        self.assertEqual(result, ('', 'R U'))

    def test_separator_at_end(self):
        """Should handle separator at end"""
        result = split_on_separator('R U,', ',')
        self.assertEqual(result, ('R U', ''))


class TestInvertMoves(unittest.TestCase):

    @patch('cubing_algs.commutator_conjugate.Algorithm')
    @patch('cubing_algs.commutator_conjugate.mirror_moves')
    def test_invert_moves(self, mock_mirror_moves, mock_algorithm_class):
        """
        Should create algorithm, transform with mirror_moves,
        and return string
        """
        # Setup mocks
        mock_algorithm_instance = MagicMock()
        mock_transformed_algo = MagicMock()
        mock_algorithm_class.return_value = mock_algorithm_instance
        mock_algorithm_instance.__add__.return_value = mock_algorithm_instance
        mock_algorithm_instance.transform.return_value = mock_transformed_algo
        mock_transformed_algo.__str__.return_value = "R' U' R U"

        # Test
        result = invert_moves("R U R' U'")

        # Assertions
        mock_algorithm_class.assert_called_once()
        mock_algorithm_instance.__add__.assert_called_once_with("R U R' U'")
        mock_algorithm_instance.transform.assert_called_once_with(mock_mirror_moves)
        self.assertEqual(result, "R' U' R U")


class TestExpandCommutatorsAndConjugates(unittest.TestCase):

    @patch('cubing_algs.commutator_conjugate.invert_moves')
    def test_simple_commutator(self, mock_invert_moves):
        """Should expand simple commutator [A, B] to A B A' B'"""
        mock_invert_moves.side_effect = lambda x: f"{x}'"

        result = expand_commutators_and_conjugates('[R U, D F]')
        expected = " R U D F R U' D F' "
        self.assertEqual(result.strip(), expected.strip())

    @patch('cubing_algs.commutator_conjugate.invert_moves')
    def test_simple_conjugate(self, mock_invert_moves):
        """Should expand simple conjugate [A: B] to A B A'"""
        mock_invert_moves.return_value = "R'"

        result = expand_commutators_and_conjugates('[R: U]')
        expected = " R U R' "
        self.assertEqual(result.strip(), expected.strip())

    @patch('cubing_algs.commutator_conjugate.invert_moves')
    def test_nested_commutator(self, mock_invert_moves):
        """Should handle nested commutators"""
        mock_invert_moves.side_effect = lambda x: f"({x})'"

        result = expand_commutators_and_conjugates('[[R, U], D]')
        # Inner commutator [R, U] expands to "R U R' U'"
        # Then outer commutator with D
        self.assertIn("R U (R)' (U)'", result)

    def test_no_brackets(self):
        """Should return unchanged string when no brackets"""
        result = expand_commutators_and_conjugates("R U R' U'")
        self.assertEqual(result, "R U R' U'")

    def test_malformed_bracket_raises_error(self):
        """Should raise InvalidBracketError for malformed brackets"""
        with self.assertRaises(InvalidBracketError) as context:
            expand_commutators_and_conjugates("[R U R'")
        self.assertIn('Malformed bracket', str(context.exception))

    def test_invalid_operator_raises_error(self):
        """Should raise InvalidOperatorError for invalid operators"""
        with self.assertRaises(InvalidOperatorError) as context:
            expand_commutators_and_conjugates('[R U | D F]')
        self.assertIn('Invalid operator', str(context.exception))

    @patch('cubing_algs.commutator_conjugate.invert_moves')
    def test_empty_bracket_parts(self, mock_invert_moves):
        """Should handle empty bracket parts"""
        mock_invert_moves.return_value = ''

        result = expand_commutators_and_conjugates('[, D]')
        expected = '  D  '
        self.assertEqual(result.strip(), expected.strip())

    @patch('cubing_algs.commutator_conjugate.invert_moves')
    def test_multiple_brackets_same_level(self, mock_invert_moves):
        """Should handle multiple brackets at same level"""
        mock_invert_moves.side_effect = lambda x: f"{x}'"

        result = expand_commutators_and_conjugates('[R, U] [D, F]')
        # Should expand both commutators
        self.assertIn("R U R' U'", result)
        self.assertIn("D F D' F'", result)

    @patch('cubing_algs.commutator_conjugate.invert_moves')
    def test_mixed_operators(self, mock_invert_moves):
        """Should handle mix of commutators and conjugates"""
        mock_invert_moves.side_effect = lambda x: f"{x}'"

        result = expand_commutators_and_conjugates('[R: U] [D, F]')
        # Should have conjugate (R U R') and commutator (D F D' F')
        self.assertIn("R U R'", result)
        self.assertIn("D F D' F'", result)

    @patch('cubing_algs.commutator_conjugate.invert_moves')
    def test_recursive_expansion(self, mock_invert_moves):
        """Should recursively expand nested structures"""
        mock_invert_moves.side_effect = lambda x: f'inv({x})'

        # Test that recursive calls are made
        with patch(
                'cubing_algs.commutator_conjugate.expand_commutators_and_conjugates',
                wraps=expand_commutators_and_conjugates,
        ) as mock_expand:
            expand_commutators_and_conjugates('[[R, U]: D]')
            # Should make multiple recursive calls
            self.assertGreater(mock_expand.call_count, 1)


class TestEdgeCases(unittest.TestCase):

    def test_empty_string(self):
        """Should handle empty strings gracefully"""
        self.assertIsNone(find_innermost_brackets(''))
        self.assertIsNone(split_on_separator('', ','))
        self.assertEqual(expand_commutators_and_conjugates(''), '')

    def test_whitespace_handling(self):
        """Should handle whitespace in brackets"""
        result = find_innermost_brackets('[ R U ]')
        self.assertEqual(result, (0, 6))

        result = split_on_separator(' R U , D F ', ',')
        self.assertEqual(result, (' R U ', ' D F '))

    @patch('cubing_algs.commutator_conjugate.invert_moves')
    def test_single_character_moves(self, mock_invert_moves):
        """Should handle single character moves"""
        mock_invert_moves.side_effect = lambda x: f"{x}'"

        result = expand_commutators_and_conjugates('[R, U]')
        expected = " R U R' U' "
        self.assertEqual(result.strip(), expected.strip())

    def test_deeply_nested_brackets(self):
        """Should handle deeply nested bracket structures"""
        nested = '[[[[[A]]]]]'
        result = find_innermost_brackets(nested)
        self.assertEqual(result, (4, 6))  # Innermost [A]



import unittest

from cubing_algs.move import Move
from cubing_algs.patterns import PATTERNS
from cubing_algs.patterns import get_pattern


class PatternsTestCase(unittest.TestCase):

    def test_patterns_size(self):
        self.assertEqual(
            len(PATTERNS.keys()),
            69,
        )

    def test_get_pattern(self):
        pattern = get_pattern('DontCrossLine')

        self.assertEqual(
            len(pattern), 6,
        )

        for m in pattern:
            self.assertTrue(isinstance(m, Move))

    def test_get_pattern_inexistant(self):
        pattern = get_pattern('El Matadore')

        self.assertEqual(
            len(pattern), 0,
        )



import unittest

from cubing_algs.parsing import parse_moves
from cubing_algs.transform.optimize import optimize_do_undo_moves
from cubing_algs.transform.optimize import optimize_double_moves
from cubing_algs.transform.optimize import optimize_repeat_three_moves
from cubing_algs.transform.optimize import optimize_triple_moves


class TransformOptimizeTestCase(unittest.TestCase):

    def test_optimize_repeat_three_moves(self):
        provide = parse_moves('R R R')
        expect = parse_moves("R'")

        self.assertEqual(
            optimize_repeat_three_moves(provide),
            expect,
        )

        provide = parse_moves("R' R' R'")
        expect = parse_moves('R')

        self.assertEqual(
            optimize_repeat_three_moves(provide),
            expect,
        )

        provide = parse_moves('R R R U')
        expect = parse_moves("R' U")

        self.assertEqual(
            optimize_repeat_three_moves(provide),
            expect,
        )

        provide = parse_moves("R' R' R' U F")
        expect = parse_moves('R U F')

        self.assertEqual(
            optimize_repeat_three_moves(provide),
            expect,
        )

        provide = parse_moves("2R' 2R' 2R' U F")
        expect = parse_moves('2R U F')

        self.assertEqual(
            optimize_repeat_three_moves(provide),
            expect,
        )

        provide = parse_moves("2R'@100 2R'@200 2R'@300 U@400 F@500")
        expect = parse_moves('2R@300 U@400 F@500')

        self.assertEqual(
            optimize_repeat_three_moves(provide),
            expect,
        )

        provide = parse_moves("2R'@100 .@150 2R'@200 2R'@300 U@400 F@500")
        expect = parse_moves("2R'@100 .@150 2R'@200 2R'@300 U@400 F@500")

        self.assertEqual(
            optimize_repeat_three_moves(provide),
            expect,
        )

        provide = parse_moves("U F R' R' R' U F")
        expect = parse_moves('U F R U F')

        self.assertEqual(
            optimize_repeat_three_moves(provide),
            expect,
        )

        self.assertEqual(
            optimize_repeat_three_moves(provide, 0),
            provide,
        )

        provide = parse_moves('U . . . U')
        expect = parse_moves('U . . . U')

        self.assertEqual(
            optimize_repeat_three_moves(provide),
            expect,
        )

    def test_optimize_do_undo_moves(self):
        provide = parse_moves("R R'")
        expect = []

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

        provide = parse_moves("R' R")

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

        provide = parse_moves("R R' U")
        expect = parse_moves('U')

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

        provide = parse_moves("R' R U F")
        expect = parse_moves('U F')

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

        provide = parse_moves("2R' 2R U F")
        expect = parse_moves('U F')

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

        provide = parse_moves("2R'@100 2R@200 U@300 F@400")
        expect = parse_moves('U@300 F@400')

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

        provide = parse_moves("2R'@100 .@150 2R@200 U@300 F@400")
        expect = parse_moves("2R'@100 .@150 2R@200 U@300 F@400")

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

        provide = parse_moves("U F R' R U F")
        expect = parse_moves('U F U F')

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

        self.assertEqual(
            optimize_do_undo_moves(provide, 0),
            provide,
        )

        provide = parse_moves('U . . . U')
        expect = parse_moves('U . . . U')

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

    def test_optimize_do_undo_double_moves(self):
        provide = parse_moves("R R R' R'")
        expect = []

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

        provide = parse_moves("R' R' R R")

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

        provide = parse_moves("R R R' R' U")
        expect = parse_moves('U')

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

        provide = parse_moves("R' R' R R U F")
        expect = parse_moves('U F')

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

        provide = parse_moves("2R' 2R' 2R 2R U F")
        expect = parse_moves('U F')

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

        provide = parse_moves("2R'@100 2R'@200 2R@300 2R@400 U@500 F@600")
        expect = parse_moves('U@500 F@600')

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

        provide = parse_moves("2R'@100 2R'@200 .@250 2R@300 2R@400 U@500 F@600")
        expect = parse_moves("2R'@100 2R'@200 .@250 2R@300 2R@400 U@500 F@600")

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

        provide = parse_moves("U F R' R' R R U F")
        expect = parse_moves('U F U F')

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

    def test_optimize_do_undo_double_double_moves(self):
        provide = parse_moves('R2 R2')
        expect = []

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

        provide = parse_moves('R2 R2 U')
        expect = parse_moves('U')

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

        provide = parse_moves('R2 R2 U F')
        expect = parse_moves('U F')

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

        provide = parse_moves('2R2 2R2 U F')
        expect = parse_moves('U F')

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

        provide = parse_moves('2R2@100 2R2@200 U@300 F@400')
        expect = parse_moves('U@300 F@400')

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

        provide = parse_moves('2R2@100 .@150 2R2@200 U@300 F@400')
        expect = parse_moves('2R2@100 .@150 2R2@200 U@300 F@400')

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

        provide = parse_moves('U F R2 R2 U F')
        expect = parse_moves('U F U F')

        self.assertEqual(
            optimize_do_undo_moves(provide),
            expect,
        )

    def test_optimize_double_moves(self):
        provide = parse_moves('R R')
        expect = parse_moves('R2')

        self.assertEqual(
            optimize_double_moves(provide),
            expect,
        )

        provide = parse_moves("R' R'")
        expect = parse_moves('R2')

        self.assertEqual(
            optimize_double_moves(provide),
            expect,
        )

        provide = parse_moves('R R U')
        expect = parse_moves('R2 U')

        self.assertEqual(
            optimize_double_moves(provide),
            expect,
        )

        provide = parse_moves("R' R' U F")
        expect = parse_moves('R2 U F')

        self.assertEqual(
            optimize_double_moves(provide),
            expect,
        )

        provide = parse_moves("2R' 2R' U F")
        expect = parse_moves('2R2 U F')

        self.assertEqual(
            optimize_double_moves(provide),
            expect,
        )

        provide = parse_moves("2R'@100 2R'@200 U@300 F@400")
        expect = parse_moves('2R2@200 U@300 F@400')

        self.assertEqual(
            optimize_double_moves(provide),
            expect,
        )

        provide = parse_moves("2R'@100 .@150 2R'@200 U@300 F@400")
        expect = parse_moves("2R'@100 .@150 2R'@200 U@300 F@400")

        self.assertEqual(
            optimize_double_moves(provide),
            expect,
        )

        provide = parse_moves("2R'@100 .@150 .@175 2R'@200 U@300 F@400")
        expect = parse_moves("2R'@100 .@150 .@175 2R'@200 U@300 F@400")

        self.assertEqual(
            optimize_double_moves(provide),
            expect,
        )

        provide = parse_moves('U . . U')
        expect = parse_moves('U . . U')

        self.assertEqual(
            optimize_double_moves(provide),
            expect,
        )

        provide = parse_moves('U F R R U F')
        expect = parse_moves('U F R2 U F')

        self.assertEqual(
            optimize_double_moves(provide),
            expect,
        )

        self.assertEqual(
            optimize_double_moves(provide, 0),
            provide,
        )

    def test_optimize_double_moves_issue_1(self):
        provide = parse_moves('R R R2 F')
        expect = parse_moves('R2 R2 F')

        self.assertEqual(
            optimize_double_moves(provide),
            expect,
        )

    def test_optimize_triple_moves(self):
        provide = parse_moves('R R2')
        expect = parse_moves("R'")

        self.assertEqual(
            optimize_triple_moves(provide),
            expect,
        )

        provide = parse_moves("R' R2")
        expect = parse_moves('R')

        self.assertEqual(
            optimize_triple_moves(provide),
            expect,
        )

        provide = parse_moves('R2 R')
        expect = parse_moves("R'")

        self.assertEqual(
            optimize_triple_moves(provide),
            expect,
        )

        provide = parse_moves("R2 R'")
        expect = parse_moves('R')

        self.assertEqual(
            optimize_triple_moves(provide),
            expect,
        )

        provide = parse_moves("R' R2 U")
        expect = parse_moves('R U')

        self.assertEqual(
            optimize_triple_moves(provide),
            expect,
        )

        provide = parse_moves("R2 R' U F")
        expect = parse_moves('R U F')

        self.assertEqual(
            optimize_triple_moves(provide),
            expect,
        )

        provide = parse_moves("R2 R' U F F")
        expect = parse_moves('R U F F')

        self.assertEqual(
            optimize_triple_moves(provide),
            expect,
        )

        provide = parse_moves("2R2 2R' U F F")
        expect = parse_moves('2R U F F')

        self.assertEqual(
            optimize_triple_moves(provide),
            expect,
        )

        provide = parse_moves("2R2@100 2R'@200 U@300 F@400 F@500")
        expect = parse_moves('2R@200 U@300 F@400 F@500')

        self.assertEqual(
            optimize_triple_moves(provide),
            expect,
        )

        provide = parse_moves("2R'@100 2R2@200 U@300 F@400 F@500")
        expect = parse_moves('2R@100 U@300 F@400 F@500')

        self.assertEqual(
            optimize_triple_moves(provide),
            expect,
        )

        provide = parse_moves("2R'@100 .@150 2R2@200 U@300 F@400 F@500")
        expect = parse_moves("2R'@100 .@150 2R2@200 U@300 F@400 F@500")

        self.assertEqual(
            optimize_triple_moves(provide),
            expect,
        )

        provide = parse_moves('U F R2 R U F')
        expect = parse_moves("U F R' U F")

        self.assertEqual(
            optimize_triple_moves(provide),
            expect,
        )

        self.assertEqual(
            optimize_triple_moves(provide, 0),
            provide,
        )

        provide = parse_moves('U . . . U')
        expect = parse_moves('U . . . U')

        self.assertEqual(
            optimize_triple_moves(provide),
            expect,
        )
